"""
Interface for Ghidra operands.
"""
from __future__ import annotations

from enum import IntFlag
from typing import TYPE_CHECKING, Optional, Tuple

from dragodis.exceptions import NotExistError
from dragodis.ghidra.variable import GhidraVariable, GhidraGlobalVariable
from dragodis.interface.operand import Operand, OperandType, ARMOperand, x86Operand
from dragodis.ghidra.operand_value import (
    OperandValue, GhidraRegister, GhidraImmediate, GhidraMemoryReference,
    GhidraRegisterList, GhidraPhrase,
)

# Used for typing.
if TYPE_CHECKING:
    from dragodis.ghidra.flat import GhidraFlatAPI
    from dragodis.ghidra.instruction import GhidraInstruction
    import ghidra
    import ghidra.app.plugin.processors.generic.Operand


# Stolen from ghidra.program.model.lang.OperandType
class GhidraOperandType(IntFlag):
    READ = 0x1
    WRITE = 0x2
    INDIRECT = 0x4
    IMMEDIATE = 0x8
    RELATIVE = 0x10
    IMPLICIT = 0x20
    CODE = 0x40
    DATA = 0x80
    PORT = 0x100
    REGISTER = 0x200
    LIST = 0x400
    FLAG = 0x800
    TEXT = 0x1000
    ADDRESS = 0x2000
    SCALAR = 0x4000
    BIT = 0x8000
    BYTE = 0x10000
    WORD = 0x20000
    QUADWORD = 0x40000
    SIGNED = 0x80000
    FLOAT = 0x100000
    COP = 0x200000
    DYNAMIC = 0x400000


class GhidraOperand(Operand):

    def __init__(self, ghidra: GhidraFlatAPI, instruction: GhidraInstruction, index: int):
        super().__init__(instruction)
        self._ghidra = ghidra
        # Ghidra doesn't necessarily have an "Operand" type.
        # Characteristics are pulled from the originating Instruction object.
        self._instruction = instruction._instruction
        self._index = index
        # TODO: Ignore validation in interest of speed?
        if index >= self._instruction.getNumOperands():
            raise NotExistError(f"The instruction at {hex(self.address)} as no operand at {index}")

    @property
    def address(self) -> int:
        return self._instruction.getAddress().getOffset()

    @property
    def index(self) -> int:
        return self._index

    # TODO: switch all these to a @cached_property
    @property
    def text(self) -> str:
        # TODO: For pointers, the text represented is the address and not the name as seen
        #   in the GUI.
        #   Figure out how to get a "OperandFieldLocation" object so we can
        #   call .getOperandRepresentation() and get the real representation.
        return str(self._instruction.getDefaultOperandRepresentation(self.index))

    @property
    def type(self) -> OperandType:
        op_type = GhidraOperandType(self._instruction.getOperandType(self.index))

        if GhidraOperandType.DYNAMIC in op_type:
            return OperandType.phrase

        if GhidraOperandType.REGISTER in op_type:
            # If also an ADDRESS, we need to check if there is an extra offset.
            # In which case it should be a phrase instead.
            # e.g. [sp,#local_4]!
            if GhidraOperandType.ADDRESS in op_type and len(self._instruction.getOpObjects(self.index)) > 1:
                return OperandType.phrase
            # In some cases, the operand will just be reported as a register even though it should
            # be interpreted as a dynamic register.
            # Check this by looking for brackets in the text.
            # e.g. the second operand should be seen as a phrase: LEA  ESP=>local_150,[ESP]
            elif (text := self.text) and "[" in text and "]" in text:
                return OperandType.phrase
            else:
                return OperandType.register

        if GhidraOperandType.ADDRESS in op_type:
            if GhidraOperandType.CODE in op_type:
                return OperandType.code
            elif GhidraOperandType.DATA in op_type:
                return OperandType.memory

        # TODO: It doesn't seem like Ghidra ever uses IMMEDIATE.
        #   So we are ignoring it to see if it ever shows up in an error log.
        if GhidraOperandType.SCALAR in op_type:
            return OperandType.immediate

        raise TypeError(f"Unexpected Ghidra OperandType {op_type} for {self.text}")

    @property
    def value(self) -> OperandValue:
        ops = self._instruction.getOpObjects(self.index)

        if not ops:
            # Some instructions return a string instead of an int for immediate values
            # e.g SAR EAX, 1
            if self.text.isdigit():
                return GhidraImmediate(int(self.text))
            else:
                raise ValueError(f"Unknown operand type: {self!r}")

        from ghidra.program.model.lang import Register
        from ghidra.program.model.address import GenericAddress
        from ghidra.program.model.scalar import Scalar

        op_types = [op.getClass() for op in ops]

        # EAX
        if op_types == [Register]:
            # A single register can be a memory access (phrase).
            # e.g. dword ptr [EAX]
            reg = GhidraRegister(ops[0])
            if self.type == OperandType.phrase:
                return GhidraPhrase(base=reg)
            else:
                return reg

        # 0x11
        elif op_types == [Scalar]:
            if self.type == OperandType.memory:
                return GhidraMemoryReference(ops[0].getValue())
            else:
                return GhidraImmediate(ops[0].getValue())

        # LAB_1000101c
        elif op_types == [GenericAddress]:
            return GhidraMemoryReference(ops[0].getOffset())

        # [EBP + 0x8]
        elif op_types == [Register, Scalar]:
            # Ghidra will also have this pattern for registers with a shift.
            # Account for this by checking type.
            if self.type == OperandType.register:
                return GhidraRegister(ops[0])
            else:
                return GhidraPhrase(
                    base=GhidraRegister(ops[0]),
                    offset=ops[1].getSignedValue(),
                )

        # [ECX + EDX*0x1]
        elif op_types == [Register, Register, Scalar]:
            return GhidraPhrase(
                base=GhidraRegister(ops[0]),
                index=GhidraRegister(ops[1]),
                scale=ops[2].getValue(),
            )

        # [ECX + EDX*0x1 + 0xf]
        elif op_types == [Register, Register, Scalar, Scalar]:
            return GhidraPhrase(
                base=GhidraRegister(ops[0]),
                index=GhidraRegister(ops[1]),
                scale=ops[2].getValue(),
                offset=ops[3].getSignedValue(),
            )

        # [r3,r2]
        # FS:[EAX]
        elif op_types == [Register, Register] and self.type != OperandType.register_list:
            return GhidraPhrase(
                base=GhidraRegister(ops[0]),
                offset=GhidraRegister(ops[1]),
            )

        # sp!,{r11 lr}
        elif all(op_type == Register for op_type in op_types):
            return GhidraRegisterList([GhidraRegister(op) for op in ops])

        # TODO: Perhaps "Phrase" should just be part of "MemoryReference"?
        # [EAX*0x4 + DAT_1000b380]
        # dword_1000B380[eax*4]
        elif op_types == [Register, Scalar, Scalar]:
            return GhidraPhrase(
                index=GhidraRegister(ops[0]),
                scale=ops[1].getValue(),
                offset=ops[2].getValue(),
            )

        raise ValueError(f"Unknown operand types: {op_types!r}")

    def _get_referenced_data_size(self, text: str) -> int:
        """
        If operand text is in a "* ptr [..]" format pull the data type name and use that to get the true size.
        """
        data_type_name = text.split()[0]
        if data_type_name == "xmmword":  # xmmword isn't an actual data type.
            return 16
        manager = self._ghidra._program.getDataTypeManager()
        data_type = manager.getDataType(f"/{data_type_name}")
        if not data_type:
            # If we fail to get the data type, try again using the builtins.
            from ghidra.program.model.data import BuiltInDataTypeManager
            manager = BuiltInDataTypeManager.getDataTypeManager()
            data_type = manager.getDataType(f"/{data_type_name}")
            if not data_type:
                raise TypeError(f"Failed to determine referenced data type for operand {self.text}")
        return data_type.getLength()

    @property
    def width(self) -> int:
        op_type = self.type

        if op_type == OperandType.register:
            # We can't call getRegister() here because Ghidra treats registers
            # with shifts to have 2 op objects [Register, Scalar], making
            # getRegister() return None.
            reg = self.value
            return reg.bit_width // 8

        elif op_type in (OperandType.immediate, OperandType.memory, OperandType.code):
            scalar = self._instruction.getScalar(self.index)
            if scalar:
                return scalar.bitLength() // 8

            # We may have a memory address which is a pointer to a standard data type.
            # Pull out the data type's size.
            if op_type == OperandType.memory:
                text = self.text
                if "ptr" in text:
                    return self._get_referenced_data_size(text)

            # Getting standard size of an address.
            addr = self._instruction.getAddress()
            if addr:
                return addr.getSize() // 8

        elif op_type == OperandType.phrase:
            # First determine if operand text is in a "* ptr [..]" format.
            # If so, pull the data type name and use that to get true size.
            # TODO: Figure out if there is a more sane way to do this.
            text = self.text
            if "ptr" in text:
                return self._get_referenced_data_size(text)
            else:
                value = self.value
                base_reg = value.base
                # Account for FS/GS pointers.
                if base_reg.name == "fs":
                    return 4
                elif base_reg.name == "gs":
                    return 8
                else:
                    return base_reg.bit_width // 8

        raise TypeError(f"Unable to get size for operand {self.text} with type: {op_type!r}")

    @property
    def variable(self) -> Optional[GhidraVariable]:
        for ref in self._ghidra._flatapi.getReferencesFrom(self._instruction.getAddress()):
            if ref.isOperandReference() and ref.getOperandIndex() == self.index:
                if ref.isStackReference():
                    stack_offset = ref.getStackOffset()
                    function = self._ghidra.get_function(self.address)
                    stack_frame = function.stack_frame
                    try:
                        return stack_frame[stack_offset]
                    except KeyError:
                        continue
                else:
                    data = self._ghidra._listing.getDataAt(ref.getToAddress())
                    if data:
                        return GhidraGlobalVariable(self._ghidra, data)


class GhidraARMOperand(GhidraOperand, ARMOperand):

    @property
    def type(self) -> OperandType:
        op_type = GhidraOperandType(self._instruction.getOperandType(self.index))

        # If we have a shift in the operand, then this should actually be a
        # register type.
        # Ghidra lazily just calls this type a pure "DYNAMIC".
        if op_type == GhidraOperandType.DYNAMIC:
            text = self.text
            if "[" not in text and any(shift_op in text for shift_op in [
                "asr", "lsl", "lsr", "ror", "rrx", "msl",
                "uxtb", "uxth", "uxtw", "uxtx",
                "sxtb", "sxth", "sxtw", "sxtx",
            ]):
                # TODO: Should we have a shifted_register operand type?
                return OperandType.register

            # If we have {} in the operand text then it is a register list.
            if "{" in text:
                return OperandType.register_list

        # Ghidra has trouble telling the difference between a [0x1234] in ARM vs
        # a 0x1234 in x86. Both are considered as SCALAR + ADDRESS.
        # While technically true, it makes it difficult to tell if we are suppose to
        # dereference it when accessing the operand value.
        # In ARM, if we have a scalar with an address we should treat it as memory.
        # In x86, the base class will treat this as an immediate.
        if op_type == GhidraOperandType.SCALAR | GhidraOperandType.ADDRESS:
            return OperandType.memory

        return super().type

    @property
    def width(self) -> int:
        """
        For ARM, we need to look at the instruction mnemonic to determine the
        predetermined size.
        """
        op_type = self.type

        # Account for size suffix in instruction when pulling phrases.
        if op_type == OperandType.phrase:
            mnemonic = self.instruction.mnemonic
            if len(mnemonic) > 1:
                if mnemonic.endswith("b"):
                    return 1
                elif mnemonic.endswith("h"):
                    return 2

        return super().width


class Ghidrax86Operand(GhidraOperand, x86Operand):
    ...



"""
Function Stack
"""

import abc
from collections.abc import MutableMapping
from typing import Union, Iterable

from dragodis.interface.variable import StackVariable


class StackFrame(MutableMapping, metaclass=abc.ABCMeta):
    """Function Stack Frame"""

    def __str__(self) -> str:
        return str(dict(self))

    def __repr__(self):
        return f"<StackFrame {self}>"

    @abc.abstractmethod
    def __getitem__(self, name_or_offset: Union[str, int]) -> StackVariable:
        """
        Obtain a stack variable from given name or offset within the stack.
        """

    def __setitem__(self, key, value):
        raise KeyError("StackFrame is read-only.")

    @abc.abstractmethod
    def __delitem__(self, name_or_offset: Union[str, int]):
        """
        Remove a stack variable from given name or offset within the stack.
        """

    @abc.abstractmethod
    def __iter__(self) -> Iterable[StackVariable]:
        """
        Iterates variable objects in the stack.
        """

    @abc.abstractmethod
    def __len__(self) -> int:
        """
        Number of variables in the stack
        """




from __future__ import annotations
from typing import Iterable, TYPE_CHECKING

from dragodis.interface import Flowchart, BasicBlock, FlowType

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI
    import ida_gdl


class IDABasicBlock(BasicBlock):

    def __init__(self, ida: IDAFlatAPI, block: "ida_gdl.BasicBlock"):
        super().__init__(ida)
        self._ida = ida
        self._block = block

    @property
    def start(self) -> int:
        return self._block.start_ea

    @property
    def end(self) -> int:
        return self._block.end_ea

    @property
    def flow_type(self) -> FlowType:
        # IDA leaves self._block.type much to be desired,
        # so we'll just look at the last instruction instead.
        for line in self.lines(reverse=True):
            return line.instruction.flow_type
        raise ValueError(f"Block at {hex(self.start)} has no instructions.")

    @property
    def flowchart(self) -> "IDAFlowchart":
        return IDAFlowchart(self._ida, self._block._fc)

    @property
    def blocks_to(self) -> Iterable["IDABasicBlock"]:
        for block in self._block.preds():
            yield IDABasicBlock(self._ida, block)

    @property
    def blocks_from(self) -> Iterable["IDABasicBlock"]:
        for block in self._block.succs():
            yield IDABasicBlock(self._ida, block)


class IDAFlowchart(Flowchart):

    def __init__(self, ida: IDAFlatAPI, flowchart: "ida_gdl.FlowChart"):
        self._ida = ida
        self._flowchart = flowchart

    def __len__(self) -> int:
        return self._flowchart.size

    @property
    def blocks(self) -> Iterable[IDABasicBlock]:
        for block in self._flowchart:
            yield IDABasicBlock(self._ida, block)




from __future__ import annotations

import abc
from typing import Optional, TYPE_CHECKING, Union, Tuple

import capstone

from dragodis.interface.types import OperandType, ARMShiftType

if TYPE_CHECKING:
    from dragodis.interface.instruction import Instruction
    from dragodis.interface.operand_value import OperandValue, Register
    from dragodis.interface.variable import Variable


class Operand(metaclass=abc.ABCMeta):
    """
    Operands represent the actual operands of instructions. Operand objects
    contain information about the important attributes of operands such as type
    and value.
    """

    def __init__(self, instruction: Instruction):
        """
        :param instruction: The underlying instruction for this operand.
        """
        self.instruction = instruction

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f"<Operand 0x{self.address:08x}:{self.index} - {self.text}>"

    @property
    def _capstone_op(self) -> Union[capstone.arm.ArmOp, capstone.x86.X86Op]:
        return self.instruction._capstone_insn.operands[self.index]

    @property
    @abc.abstractmethod
    def address(self) -> int:
        """
        The address of the instruction the operand is found in.
        """

    @property
    @abc.abstractmethod
    def index(self) -> int:
        """
        Index of operand as found in the instruction.
        """

    @property
    @abc.abstractmethod
    def text(self) -> str:
        """
        Disassembled text for operand.
        """

    @property
    @abc.abstractmethod
    def type(self) -> OperandType:
        """
        Type of operand.
        """

    @property
    @abc.abstractmethod
    def value(self) -> OperandValue:
        """
        An object that represents the contents of the operand
        based on the type.

        :return: An OperandValue object.
        """

    # TODO: Rename to "size" instead?
    @property
    @abc.abstractmethod
    def width(self) -> int:
        """
        Based on the data type it contains, the size of the operand's value in bytes
        """

    @property
    @abc.abstractmethod
    def variable(self) -> Optional[Variable]:
        """
        Obtains the variable referenced within the operand.
        Returns None, if no variable is referenced.
        """

    # TODO: Add more attributes for pulling individual components for phrases.


class ARMOperand(Operand):
    """
    ARM based operand.
    """

    _capstone_shift_map = {
        capstone.arm.ARM_SFT_ASR: (ARMShiftType.ASR, False),
        capstone.arm.ARM_SFT_LSL: (ARMShiftType.LSL, False),
        capstone.arm.ARM_SFT_LSR: (ARMShiftType.LSR, False),
        capstone.arm.ARM_SFT_ROR: (ARMShiftType.ROR, False),
        capstone.arm.ARM_SFT_RRX: (ARMShiftType.RRX, False),
        capstone.arm.ARM_SFT_ASR_REG: (ARMShiftType.ASR, True),
        capstone.arm.ARM_SFT_LSL_REG: (ARMShiftType.LSL, True),
        capstone.arm.ARM_SFT_LSR_REG: (ARMShiftType.LSR, True),
        capstone.arm.ARM_SFT_ROR_REG: (ARMShiftType.ROR, True),
        capstone.arm.ARM_SFT_RRX_REG: (ARMShiftType.RRX, True),
    }

    @property
    def _capstone_op(self) -> capstone.arm.ArmOp:
        """
        Convenience function for getting capstone implementation of operand.
        """
        return self.instruction._capstone_insn.operands[self.index]

    # TODO: Should shifted information be stored in a ShiftedRegister OperandValue type instead?
    @property
    def shift(self) -> Tuple[ARMShiftType, Union[int, Register]]:
        """
        The shift type and shift amount (which can be a constant or another register)
        (Should default to LSL with a shift count of 0 for non-shifted operands.)

        Defaults to using capstone.
        """
        op = self._capstone_op

        if op.shift.type == capstone.arm.ARM_SFT_INVALID:
            return ARMShiftType.LSL, 0  # not shifted

        try:
            shift_type, is_reg = self._capstone_shift_map[op.shift.type]
        except KeyError:
            raise NotImplementedError(f"Unsupported shift type: {op.shift.type}")

        if is_reg:
            reg_name = self.instruction._capstone_insn.reg_name(op.shift.value)
            shift_count = self.instruction._api.get_register(reg_name)
        else:
            shift_count = op.shift.value

        return shift_type, shift_count


class x86Operand(Operand):
    """
    x86 based operand.
    """

    @property
    def _capstone_op(self) -> capstone.x86.X86Op:
        """
        Convenience function for getting capstone implementation of operand.
        """
        return self.instruction._capstone_insn.operands[self.index]



"""
Home for constants and types.
"""

from enum import IntEnum, IntFlag, auto
from typing import Any, List


class CommentType(IntEnum):
    eol = auto()
    anterior = auto()
    posterior = auto()
    plate = auto()
    repeatable = auto()


class FlowType(IntEnum):
    call = auto()
    terminal = auto()
    conditional_jump = auto()
    unconditional_jump = auto()
    fall_through = auto()


# TODO: We should add a type for func_ptr, (which would overwrite being DWORD or whatever)
class LineType(IntEnum):
    code = auto()
    byte = auto()
    word = auto()
    dword = auto()
    qword = auto()
    oword = auto()  # also known as xmmword (16 byte integer)
    tbyte = auto()
    float = auto()
    double = auto()
    pack_real = auto()
    string = auto()     # utf-8 string type
    string16 = auto()   # utf-16-le string type
    string32 = auto()   # utf-32-le string type
    struct = auto()
    align = auto()
    tail = auto()
    undefined = auto()
    unloaded = auto()

    @classmethod
    def match_type(cls, value: Any) -> List["LineType"]:
        """
        Returns all possible LineTypes that match to the given value.
        List is sorted so that the first entry would be default.
        """
        # Import here to prevent cyclic loop.
        from dragodis.interface import Instruction

        if isinstance(value, Instruction):
            return [cls.code]
        elif isinstance(value, int):
            types = []
            if value < 0:
                return []  # TODO: handle negative values?
            if value <= 0xffffffffffffffffffffffffffffffff:
                types.append(cls.oword)
            if value <= 0xffffffffffffffff:
                types.append(cls.qword)
            if value <= 0xffff:
                types.append(cls.word)
            if value <= 0xff:
                types.append(cls.byte)
                types.append(cls.undefined)
            if value <= 0xffffffff:  # checking last so we can have dword as default over byte and word
                types.append(cls.dword)
            types.reverse()
            return types
        elif isinstance(value, float):
            return [cls.float, cls.double]
        elif isinstance(value, str):
            return [cls.string, cls.string16, cls.string32]
        # TODO: should code count as bytes?
        elif isinstance(value, bytes) and len(value) == 1:
            return [cls.byte, cls.undefined, cls.align]
        elif isinstance(value, bytes):
            return [cls.undefined, cls.align]
        elif value is None:
            return [cls.unloaded]
        # TODO: Support this?
        # elif isinstance(value, dict):
        #     return [self.struct]
        else:
            return []


# TODO: Add description and example for each type.
class OperandType(IntEnum):
    """Types of operands."""
    void = auto()           # No Operand
    register = auto()       # General Register
    register_list = auto()  # List of Registers
    memory = auto()         # Direct Memory Reference  # TODO: Rename this to "data"?
    immediate = auto()      # Immediate Value
    phrase = auto()         # Memory Phrase/Displacement
    code = auto()           # Immediate Near/Far Address to code (usually the start of a function)


class ARMShiftType(IntEnum):
    """
    The possible shift types found in ARM operands.
    """
    LSL = 0       # logical left         LSL #0 - (don't shift)
    LSR = auto()  # logical right        LSR #0 means LSR #32
    ASR = auto()  # arithmetic right     ASR #0 means ASR #32
    ROR = auto()  # rotate right         ROR #0 means RRX
    RRX = auto()  # extended rotate right

    # ARMv8 shifts
    MSL = auto()  # masked shift left (ones are shifted in from the right)

    # extending register operations
    UXTB = auto()
    UXTH = auto()
    UXTW = auto()
    UXTX = auto()  # alias for LSL
    SXTB = auto()
    SXTH = auto()
    SXTW = auto()
    SXTX = auto()


# TODO: Define what these mean
# TODO: Should these be flags instead of exclusive?
#   - We can have both a read and a write!
class ReferenceType(IntEnum):
    """Types of references."""
    # TODO: These where originally based on IDA. Get influence from Ghidra
    unknown = auto()
    data = auto()
    data_offset = auto()  #TODO: what is this?
    data_write = auto()
    data_read = auto()
    data_text = auto()
    data_informational = auto()
    code_call = auto()
    # code_far_call = auto()
    # code_near_call = auto()
    code_jump = auto()
    # code_far_jump = auto()
    # code_near_jump = auto()
    code_user = auto()
    ordinary_flow = auto()


class SegmentPermission(IntFlag):
    execute = auto()
    write = auto()
    read = auto()
    volatile = auto()


class SegmentType(IntEnum):
    code = auto()
    data = auto()
    null = auto()
    uninitialized = auto()
    external = auto()


class ProcessorType(IntEnum):
    ARM = auto()
    ARM64 = auto()
    x86 = auto()
    x64 = auto()


# convenience
PROCESSOR_ARM = ProcessorType.ARM
PROCESSOR_ARM64 = ProcessorType.ARM64
PROCESSOR_X86 = ProcessorType.x86
PROCESSOR_X64 = ProcessorType.x64




from __future__ import annotations
from typing import TYPE_CHECKING

from dragodis.interface.data_type import DataType

if TYPE_CHECKING:
    from dragodis import IDA
    import ida_typeinf


class IDADataType(DataType):

    def __init__(
            self,
            ida: IDA,
            tinfo: "ida_typeinf.tinfo_t" = None,
            address: int = None
    ):
        if not (tinfo or address):
            raise ValueError(f"Must provide a tinfo object or address.")
        self._ida = ida
        self._tinfo = tinfo
        self._address = address

    @property
    def name(self) -> str:
        if self._tinfo:
            return str(self._tinfo).lower().strip("_")
        else:
            TYPE_MAP = {
                self._ida._idc.FF_BYTE: "byte",
                self._ida._idc.FF_WORD: "word",
                self._ida._idc.FF_DWORD: "dword",
                self._ida._idc.FF_QWORD: "qword",
                self._ida._idc.FF_OWORD: "oword",
                self._ida._idc.FF_TBYTE: "tbyte",
                self._ida._idc.FF_STRLIT: "char",
                self._ida._idc.FF_STRUCT: "struct",
                self._ida._idc.FF_FLOAT: "float",
                self._ida._idc.FF_DOUBLE: "double",
                self._ida._idc.FF_PACKREAL: "packed decimal real",
                self._ida._idc.FF_ALIGN: "alignment directive",
            }
            flags = self._ida._ida_bytes.get_flags(self._address)
            flags &= self._ida._ida_bytes.DT_TYPE
            return TYPE_MAP[flags]

    @property
    def size(self) -> int:
        if self._tinfo:
            return self._tinfo.get_size()
        else:
            flags = self._ida._ida_bytes.get_flags(self._address)
            flags &= self._ida._ida_bytes.DT_TYPE
            return self._ida._ida_bytes.get_data_elsize(self._address, flags)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Union

from dragodis.ida.variable import IDAStackVariable
from dragodis.interface.stack import StackFrame

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI
    import ida_struct


class IDAStackFrame(StackFrame):

    def __init__(self, ida: IDAFlatAPI, frame: "ida_struct.struc_t"):
        self._ida = ida
        self._frame = frame
        # Obtain the base of the stack by obtaining the location of the return address.
        func = self._ida._ida_funcs.get_func(
            self._ida._ida_frame.get_func_by_frame(self._frame.id)
        )
        self._base_offset = self._ida._ida_frame.frame_off_retaddr(func)
        self._retsize = self._ida._ida_frame.get_frame_retsize(func)

    def __eq__(self, other):
        return self is other or self._frame.id == other._frame.id

    def __getitem__(self, name_or_offset: Union[str, int]) -> IDAStackVariable:
        if isinstance(name_or_offset, int):
            offset = name_or_offset + self._base_offset
            member = self._ida._ida_struct.get_member(self._frame, offset)
        elif isinstance(name_or_offset, str):
            name = name_or_offset
            member = self._ida._ida_struct.get_member_by_name(self._frame, name)
        else:
            raise TypeError(f"Must be str or int, got {type(name_or_offset)}")
        if not member:
            raise KeyError(f"Unable to find stack variable from: {name_or_offset}")
        return IDAStackVariable(self._ida, self, member)

    def __delitem__(self, name_or_offset: Union[str, int]):
        variable = self[name_or_offset]
        self._ida._ida_struct.del_struc_member(self._frame, variable._member.soff)

    def __iter__(self) -> Iterable[IDAStackVariable]:
        for member in self._frame.members:
            variable = IDAStackVariable(self._ida, self, member)
            # Ignore the hidden " s" and " r" variables.
            # TODO: Determine meaning behind these.
            if variable.name not in (" s", " r"):
                yield variable

    def __len__(self) -> int:
        # - 1 for the " s" hidden member and - 1 if there would be a " r" hidden return
        # variable as well.
        return self._frame.memqty - 1 - (1 if self._retsize else 0)



"""
Defines interface for different types of values found within an operand.

NOTE: Some of these are interfaces that must be subclassed to be implemented based on
    the underlying disassembler.
    Others can be used directly.
"""
from __future__ import annotations
import abc
from functools import total_ordering
from typing import Optional, Union, List


class OperandValue(metaclass=abc.ABCMeta):
    """
    Base class for all possible operand value types.
    """

    def __int__(self):
        """
        Operand values can be casted to int to get a sane value based on type.
        (This is equivalent to IDA's idc.get_operand_value() function)

        WARNING: Not all OperandValue subclasses have an equivalent int value (e.g. Register),
        in which case it will return a -1.
        """
        return -1


class Immediate(int, OperandValue):
    """
    Defines an immediate or constant used in an operand.
    """

    def __str__(self) -> str:
        return str(int(self))

    def __repr__(self) -> str:
        return f"<Immediate {self}>"


class MemoryReference(int, OperandValue):
    """
    Memory reference is a special type of immediate, where the value represents a defined
    item in the disassembler.
    """

    def __str__(self) -> str:
        return f"0x{self:08x}"

    def __repr__(self) -> str:
        return f"<MemoryReference {self}>"


@total_ordering
class Register(OperandValue, metaclass=abc.ABCMeta):
    """
    Register objects represent the register components of operands.
    """

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Register {self.name} - bit_width={self.bit_width}>"

    @abc.abstractmethod
    def __eq__(self, register: Register):
        ...

    def __lt__(self, other: Register):
        return (self.base.name, self.mask) < (other.base.name, other.mask)

    @property
    @abc.abstractmethod
    def base(self) -> Register:
        """The full size register in this register's family."""

    @property
    @abc.abstractmethod
    def bit_width(self) -> int:
        """The total number of bits for this register."""

    @property
    @abc.abstractmethod
    def mask(self) -> int:
        """A mask indicating the applicable bits for this register within the base register."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The name of the register."""


class RegisterList(List[Register], OperandValue):
    """
    Defines a list of registers used as an operand.

    .. code::

        {R4-R10,LR}
    """

    def __str__(self) -> str:
        return f"{{{','.join(self)}}}"

    def __repr__(self) -> str:
        return f"<RegisterList {super().__repr__()}>"


class Phrase(OperandValue):
    """
    Defines an operand phrase of one of the following forms:

    .. code::

        [base + index * scale + offset]
        [base + offset]
    """

    # TODO: For capstone, x86 had "segment" and ARM had "lshift"

    def __int__(self):
        return self.offset

    def __str__(self) -> str:
        segments = []
        if (base := self.base) is not None:
            segments.append(str(base))
        if (index := self.index) is not None:
            segments.append(f"{index}*0x{self.scale:x}")
        if offset := self.offset:
            if isinstance(offset, int):
                segments.append(f"0x{offset:x}")
            else:
                segments.append(str(offset))
        return f"[{' + '.join(segments)}]"

    def __repr__(self) -> str:
        offset = self.offset
        if isinstance(offset, int):
            offset = f"0x{offset:x}"
        return f"<Phrase base={self.base}, index={self.index}, scale={self.scale}, offset={offset}>"

    @property
    @abc.abstractmethod
    def base(self) -> Optional[Register]:
        """
        The base register if operand is a phrase.
        May be None if there is no base:

        .. code::

            dword ptr [EAX*0x4 + DAT_0040dc20]
        """

    @property
    @abc.abstractmethod
    def index(self) -> Optional[Register]:
        """
        The index register
        (May be None if phrase doesn't have scaled index.)
        """

    @property
    @abc.abstractmethod
    def scale(self) -> int:
        """
        The scaling factor for the index.
        NOTE: This should default to 0 if index * scale is not supported in the processor. (ARM)
        """

    @property
    @abc.abstractmethod
    def offset(self) -> Union[Register, int]:
        """
        The offset or displacement.
        This could be a register or immediate.
        """







from __future__ import annotations
from typing import Tuple, TYPE_CHECKING

from dragodis.ida.operand_value import IDARegister
from dragodis.interface.function_argument_location import (
    ArgumentLocation, StackLocation, RegisterLocation,
    RegisterPairLocation, RelativeRegisterLocation, StaticLocation
)

if TYPE_CHECKING:
    import ida_typeinf
    from dragodis.ida.flat import IDAFlatAPI


class IDAArgumentLocation(ArgumentLocation):

    def __init__(self, ida: IDAFlatAPI, argloc: "ida_typeinf.argloc_t", size: int):
        self._ida = ida
        self._argloc = argloc
        self._size = size


class IDAStackLocation(StackLocation, IDAArgumentLocation):

    @property
    def stack_offset(self) -> int:
        return self._argloc.stkoff()


class IDARegisterLocation(RegisterLocation, IDAArgumentLocation):

    @property
    def register(self) -> IDARegister:
        return IDARegister(self._ida, self._argloc.reg1(), self._size)


class IDARegisterPairLocation(RegisterPairLocation, IDAArgumentLocation):

    @property
    def registers(self) -> Tuple[IDARegister, IDARegister]:
        # Size is the combination of both registers.
        size = self._size // 2
        return (
            IDARegister(self._ida, self._argloc.reg1(), size),
            IDARegister(self._ida, self._argloc.reg2(), size),
        )


class IDARelativeRegisterLocation(RelativeRegisterLocation, IDARegisterLocation):

    @property
    def register(self) -> IDARegister:
        rrel = self._argloc.get_rrel()
        return IDARegister(self._ida, rrel.reg, self._size)

    @property
    def offset(self) -> int:
        rrel = self._argloc.get_rrel()
        return rrel.off


class IDAStaticLocation(StaticLocation, IDAArgumentLocation):

    @property
    def address(self) -> int:
        return self._argloc.get_ea()





from __future__ import annotations

from dragodis.ida.data_type import IDADataType
from dragodis.interface.variable import StackVariable, GlobalVariable
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI
    from dragodis.ida.stack import IDAStackFrame
    import ida_struct


class IDAGlobalVariable(GlobalVariable):

    def __init__(self, ida: IDAFlatAPI, address: int):
        self._ida = ida
        self._address = address

    @property
    def address(self) -> int:
        return self._address

    @property
    def name(self) -> str:
        return self._ida.get_name(self._address)

    @name.setter
    def name(self, new_name: str):
        self._ida.set_name(self._address, new_name)

    @property
    def size(self) -> int:
        return self._ida._ida_bytes.get_item_size(self._address)

    @property
    def data_type(self) -> IDADataType:
        return IDADataType(self._ida, address=self._address)


class IDAStackVariable(StackVariable):

    def __init__(self, ida: IDAFlatAPI, frame: IDAStackFrame, member: "ida_struct.member_t"):
        self._ida = ida
        self._frame = frame
        self._member = member

    def __eq__(self, other: "IDAVariable") -> bool:
        return self._frame == other._frame and self._member.id == other._member.id

    @property
    def stack_offset(self) -> int:
        return self._member.soff - self._frame._base_offset

    @property
    def name(self) -> str:
        return self._ida._ida_struct.get_member_name(self._member.id) or ""

    @name.setter
    def name(self, new_name: str):
        self._ida._ida_struct.set_member_name(self._frame._frame, self._member.soff, new_name)

    @property
    def data_type(self) -> IDADataType:
        tif = self._ida._ida_typeinf.tinfo_t()
        success = self._ida._ida_struct.get_or_guess_member_tinfo(tif, self._member)
        if not success:
            raise RuntimeError("Unexpected error getting type information.")
        return IDADataType(self._ida, tif)  #, self._member.flag & self._ida._idc.DT_TYPE)

    @property
    def size(self) -> int:
        return self._ida._ida_struct.get_member_size(self._member)


IDAVariable = Union[IDAGlobalVariable, IDAStackVariable]




from __future__ import annotations
import logging
import re
from typing import TYPE_CHECKING, List, Union, Optional

from dragodis.ida.data_type import IDADataType
from dragodis.ida.function_argument_location import (
    IDAArgumentLocation, IDAStaticLocation, IDARegisterLocation,
    IDAStackLocation, IDARegisterPairLocation, IDARelativeRegisterLocation
)
from dragodis.interface.function_signature import FunctionSignature, FunctionParameter
from dragodis.exceptions import NotExistError, UnsupportedError

if TYPE_CHECKING:
    import ida_typeinf
    from dragodis.ida.flat import IDAFlatAPI

logger = logging.getLogger(__name__)


class IDAFunctionSignature(FunctionSignature):

    # pulled from ida_typeinf.CM_CC_* constants
    _cc_map = {
        0x10: "__unknown",
        0x20: "__voidarg",
        0x30: "__cdecl",
        0x40: "__ellipsis",
        0x50: "__stdcall",
        0x60: "__Pascal",
        0x70: "__fastcall",
        0x80: "__thiscall",
        0xB0: "__golang",
        0xD0: "__usercall",
        0xE0: "__userpurge",
        0xF0: "__usercall",
    }
    # _cc_map_inv will end up with __usercall as 0xf0, which is the more sensible default
    _cc_map_inv = {name: opcode for opcode, name in _cc_map.items()}

    def __init__(self, ida: IDAFlatAPI, address: int):
        self._address = address
        self._ida = ida
        try:
            self._func_type_data, self._tif = ida._ida_helpers.get_func_type_info(address)
        except RuntimeError:
            raise NotExistError(f"A function signature does not exist at 0x{address:08X}")
        self._parameters: Optional[List[IDAFunctionParameter]] = None

    @property
    def name(self) -> str:
        idc = self._ida._idc

        # (Using get_name() over get_func_name() so it also works for imported functions)
        func_name = idc.get_name(self._address)
        # Demangle name if necessary:
        demangled_name = idc.demangle_name(func_name, idc.get_inf_attr(idc.INF_SHORT_DN))
        if demangled_name:
            # Strip off the extra junk: 'operator new(uint)' -> 'new'
            match = re.search(r"([^ :]*)\(", demangled_name)
            if not match:
                logger.debug(f"Unable to demangle function name: {demangled_name}")
            else:
                logger.debug(f"Demangled function name {demangled_name} -> {match.group(1)}")
                demangled_name = match.group(1)
            func_name = demangled_name
        return func_name

    @property
    def declaration(self) -> str:
        # There is no way to get the full function declaration directly from IDA with
        # the function name intact. So we have to recreate it.
        # If function doesn't have a name (usually because the function was dynamically created
        # within a register), then we are just going to call it "no_name" so we can still get the
        # function typing to still work.
        return re.sub(r'\(', f' {self.name or "no_name"}(', f'{str(self._tif)};')

    @property
    def calling_convention(self) -> str:
        cc = self._func_type_data.cc & self._ida._ida_typeinf.CM_CC_MASK
        try:
            return self._cc_map[cc]
        except KeyError:
            raise RuntimeError(f"{self} has unexpected calling convention: {hex(cc)}")

    @calling_convention.setter
    def calling_convention(self, name: str):
        if not name.startswith("__"):
            name = f"__{name}"
        name = name.lower()
        try:
            cc = self._cc_map_inv[name]
        except KeyError:
            raise ValueError(f"Invalid calling convention name: {name}")
        # Set calling convention part of cm_t flags.
        cc |= self._func_type_data.cc & (self._ida._ida_typeinf.CM_CC_MASK ^ 0xff)
        self._func_type_data.cc = cc
        self._apply()
        self._parameters = None

    @property
    def return_type(self) -> IDADataType:
        return IDADataType(self._ida, self._func_type_data.rettype)

    @return_type.setter
    def return_type(self, data_type: Union[IDADataType, str]):
        if isinstance(data_type, str):
            data_type = self._ida.get_data_type(data_type)
        self._func_type_data.rettype = data_type._tinfo
        self._apply()

    @property
    def parameters(self) -> List[IDAFunctionParameter]:
        if self._parameters is None:
            self._parameters = [
                IDAFunctionParameter(self, i) for i in range(self._func_type_data.size())
            ]
            if len(self._parameters) <= 1:
                return self._parameters
            if self.calling_convention in ("__usercall", "__userpurge"):
                self._parameters = sorted(self._parameters, key=lambda parameter: parameter.location)
        return self._parameters

    def _apply(self):
        """
        Applies changes to func_type_data back to the IDB database.
        """
        self._func_type_data, self._tif = self._ida._ida_helpers.apply_func_type_data(self._address, self._func_type_data)

    def replace_parameters(self, data_types: List[str]):
        # Easiest way to replace all parameters is to set the function type itself.
        # Since IDA doesn't give us names, we are just going to name the arguments ourselves
        # as "a1" where 1 is the index of the parameter.
        parameters = [f"{data_type} a{i}" for i, data_type in enumerate(data_types, start=1)]
        declaration = re.sub(r"\(.*\)", f"({','.join(parameters)})", self.declaration)

        ida_typeinf = self._ida._ida_typeinf
        tif = ida_typeinf.tinfo_t()
        til = ida_typeinf.get_idati()
        func_type_data = ida_typeinf.func_type_data_t()
        ida_typeinf.parse_decl(tif, til, declaration, ida_typeinf.PT_SIL)
        tif.get_func_details(func_type_data)
        self._tif = tif
        self._func_type_data = func_type_data
        self._apply()

        # Reset cache
        self._parameters = None

    def add_parameter(self, data_type: str) -> IDAFunctionParameter:
        # Create new parameter in func_type_data
        data_type = self._ida.get_data_type(data_type)
        new_param = self._ida._ida_typeinf.funcarg_t()
        new_param.type = data_type._tinfo
        self._func_type_data.push_back(new_param)
        self._apply()

        # Update _parameters cache.
        if self._parameters is not None:
            new_param = IDAFunctionParameter(self, len(self._parameters))
            self._parameters.append(new_param)
            return new_param
        else:
            return self.parameters[-1]

    def remove_parameter(self, ordinal: int):
        num_parameters = self._func_type_data.size()
        if ordinal < 0:
            ordinal += num_parameters
        if ordinal not in range(num_parameters):
            raise NotExistError(f"Parameter doesn't exist at ordinal: {ordinal}")
        self._func_type_data.erase(self._func_type_data[ordinal])
        self._apply()

        # Update parameters cache.
        if self._parameters is not None:
            del self._parameters[ordinal]

    def insert_parameter(self, ordinal: int, data_type: str) -> IDAFunctionParameter:
        num_parameters = self._func_type_data.size()
        if ordinal < 0:
            ordinal += num_parameters
        if ordinal not in range(num_parameters):
            raise ValueError(f"Invalid ordinal for parameter insertion: {ordinal}")

        if self.calling_convention in ("__usercall", "__userpurge"):
            raise UnsupportedError("Cannot insert a parameter into a usercall function")

        data_type = self._ida.get_data_type(data_type)
        new_param = self._ida._ida_typeinf.funcarg_t()
        new_param.type = data_type._tinfo
        # Getting an element actually gives back a funcarg_t iterator currently
        # pointing at the element, which is what we want to be able to use the
        # C++ vector object.
        it = self._func_type_data[ordinal]
        self._func_type_data.insert(it, new_param)
        self._apply()

        # Update parameters cache.
        if self._parameters is not None:
            new_param = IDAFunctionParameter(self, ordinal)
            self._parameters.insert(ordinal, new_param)
            for i, param in enumerate(self._parameters):
                param._ordinal = i
            return new_param
        else:
            return self.parameters[ordinal]


class IDAFunctionParameter(FunctionParameter):
    """
    Interface for a parameter from a FunctionSignature
    """

    def __init__(self, signature: IDAFunctionSignature, ordinal: int):
        # NOTE: While it may look weird that there is no argument specific stuff here, that
        # is because the signature will keep track of the ordinal for us, which gives
        # us the information to obtain everything else.
        # This allows us to dynamically change the ordinal of the parameter based on signature.
        super().__init__(signature)
        self._ida = signature._ida
        self._ordinal = ordinal

    @property
    def _funcarg(self) -> "ida_typeinf.funcarg_t":
        return self.signature._func_type_data[self.ordinal]

    @property
    def name(self) -> str:
        return self._funcarg.name

    @name.setter
    def name(self, new_name: str):
        self._funcarg.name = new_name
        self.signature._apply()

    @property
    def ordinal(self) -> int:
        return self._ordinal

    @property
    def data_type(self) -> IDADataType:
        return IDADataType(self._ida, self._funcarg.type)

    @data_type.setter
    def data_type(self, new_type: Union[str, IDADataType]):
        if isinstance(new_type, str):
            new_type = self._ida.get_data_type(new_type)
        self._funcarg.type = new_type._tinfo
        self.signature._apply()

    @property
    def declaration(self) -> str:
        return f"{self._funcarg.type} {self._funcarg.name}"

    @property
    def location(self) -> IDAArgumentLocation:
        argloc = self._funcarg.argloc
        if argloc.is_stkoff():
            return IDAStackLocation(self._ida, argloc, self.data_type.size)
        elif argloc.is_reg1():
            return IDARegisterLocation(self._ida, argloc, self.data_type.size)
        elif argloc.is_reg2():
            return IDARegisterPairLocation(self._ida, argloc, self.data_type.size)
        elif argloc.is_rrel():
            return IDARelativeRegisterLocation(self._ida, argloc, self.data_type.size)
        elif argloc.is_ea():
            return IDAStaticLocation(self._ida, argloc, self.data_type.size)
        else:
            raise ValueError(f"Unsupported argument location type: {argloc.atype()}")



"""
Stores global constants used in dragodis
"""
import os

# backend disassembler names
BACKEND_GHIDRA = "Ghidra"
BACKEND_IDA = "IDA"

_default = os.environ.get("DRAGODIS_DISASSEMBLER", "")
if _default.lower() == BACKEND_GHIDRA.lower():
    _default = BACKEND_GHIDRA
elif _default.lower() == BACKEND_IDA.lower():
    _default = BACKEND_IDA
elif _default:
    valid = [BACKEND_GHIDRA, BACKEND_IDA]
    raise ValueError(f"Invalid DRAGODIS_DISASSEMBLER set: '{_default}'. Should be one of: {','.join(valid)}")
BACKEND_DEFAULT = _default or None




from __future__ import annotations
from typing import TYPE_CHECKING

import bytesparse

from dragodis.exceptions import NotExistError
from dragodis.interface import Memory

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI


class CachedMemory:

    _PAGE_SIZE = 0x1000

    def __init__(self, ida: IDAFlatAPI):
        self._ida = ida
        # Blocks we still need to cache
        uncached_chunks = []
        for start, size in ida._ida_helpers.get_all_byte_chunks():
            # Split up large chunks into page size chunks.
            while size > self._PAGE_SIZE:
                uncached_chunks.append((start, self._PAGE_SIZE))
                start += self._PAGE_SIZE
                size -= self._PAGE_SIZE
            if size:
                uncached_chunks.append((start, size))

        self._uncached_chunks = uncached_chunks
        self._memory = bytesparse.Memory()

    def _obtain_uncached_chunks(self, address: int, size: int):
        """
        Obtains the uncached chunks that overlap the given address range,
        then pulls the chunks from IDA.
        """
        if not size:
            return

        start = address
        end = address + size

        index = 0
        while index < len(self._uncached_chunks):
            chunk_start, chunk_size = self._uncached_chunks[index]
            chunk_end = chunk_start + chunk_size
            if chunk_start >= end:
                # Chunks are in order and non-overlapping, so we can break
                # if we see a chunk past our range.
                break
            if start < chunk_end and chunk_start < end:  # overlaps?
                data = self._ida._ida_bytes.get_bytes(chunk_start, chunk_size)
                self._memory.write(chunk_start, data)
                del self._uncached_chunks[index]
            else:
                index += 1

    def get(self, address: int, size: int, fill_pattern=b"\x00") -> bytes:
        """
        Obtains bytes from given address range.
        Non-initialized bytes are filled with 0 bytes.

        :param address: Start address.
        :param size: Number of bytes to obtain.
        :param fill_pattern: byte value to fill non-initialized bytes.
            If set to None, a NotExistError will be thrown if data is not contiguous.
        """
        self._obtain_uncached_chunks(address, size)
        if fill_pattern:
            return bytes(self._memory[address:address+size:fill_pattern])
        else:
            try:
                return bytes(self._memory[address:address+size])
            except ValueError as e:
                raise NotExistError(f"Unable to obtain {size} bytes from 0x{address:08X}: {e}")

    def set(self, address: int, data: bytes):
        """
        Updates cached data with given data.
        """
        # Pull from IDB first to update cached blocks.
        # Usually we wouldn't be setting data at something we haven't
        # pulled from first, so the performance hit is acceptable.
        self._obtain_uncached_chunks(address, len(data))
        self._memory.write(address, data)

    def is_contiguous(self, address: int, size) -> bool:
        """
        Determines if the address range was fully loaded.
        Ie. there are no holes of uninitialized bytes (is_loaded() == False)
        """
        self._obtain_uncached_chunks(address, size)
        return self._memory[address:address+size].contiguous


class IDAMemory(Memory):
    """
    A file-like interface for  memory within IDA.
    Performs caching and buffering to improve performance.

    NOTE: Unloaded memory will be presented as zero bytes.
    """

    def __init__(self, ida: IDAFlatAPI, start: int, end: int):
        super().__init__(start, end)
        self._ida = ida
        self._cache = ida._cached_memory

    def read(self, size: int = None) -> bytes:
        remaining_bytes = self.end - (self.start + self._offset)
        if size is None:
            size = remaining_bytes
        size = min(size, remaining_bytes)
        if not size:
            return b""

        address = self.start + self._offset
        data = self._cache.get(address, size)

        self._offset += len(data)
        return data

    def reset(self, size: int = None) -> int:
        remaining_bytes = self.end - (self.start + self._offset)
        if size is None:
            size = remaining_bytes
        size = min(size, remaining_bytes)
        if not size:
            return 0

        address = self.start + self._offset
        self._ida._ida_helpers.revert_bytes(address, size)

        # Update cache.
        data = self._ida._ida_helpers.get_bytes(address, size)
        self._cache.set(address, data)

        self._offset += size
        return size

    def write(self, data: bytes) -> int:
        # Trim given data to ensure we only write within the window.
        data = data[:self.end - (self.start + self._offset)]
        if not data:
            return 0

        address = self.start + self._offset

        # Check if we would be writing to an uninitialized section.
        if not self._ida._ida_helpers.is_loaded(address, len(data)):
            raise IOError(f"Unable to write to address not fully initialized: 0x{address:08x}")

        # Patch bytes with given data.
        self._ida._ida_bytes.patch_bytes(address, data)

        # Update cache.
        self._cache.set(address, data)

        self._offset += len(data)
        return len(data)




from __future__ import annotations
from typing import TYPE_CHECKING

from dragodis.interface.string import String

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI
    import idautils


class IDAString(String):

    def __init__(self, ida: IDAFlatAPI, string: "idautils.Strings.StringItem"):
        self._ida = ida
        self._string = string

    @property
    def value(self) -> str:
        return str(self._string)

    @property
    def data(self) -> bytes:
        return self._ida._ida_bytes.get_strlit_contents(self._string.ea, self._string.length, self._string.strtype)

    @property
    def address(self) -> int:
        return self._string.ea




from __future__ import annotations
from typing import TYPE_CHECKING, Union, Optional

from dragodis.interface.operand_value import (
    OperandValue, Immediate, MemoryReference, Register,
    RegisterList, Phrase,
)

if TYPE_CHECKING:
    import ida_ua
    from dragodis.ida.flat import IDAFlatAPI


class IDAImmediate(Immediate):
    ...


class IDAMemoryReference(MemoryReference):
    ...


class IDARegister(Register):

    def __init__(self, ida: IDAFlatAPI, reg: int, width: int):
        """
        :param ida: The IDA disassembler.
        :param reg: Internal register number as defined in the IDA processor module.
        :param width: The size of the register in bytes.
        """
        self._ida = ida
        self._reg = reg
        self._width = width

    def __eq__(self, other: "IDARegister"):
        if isinstance(other, IDARegister):
            return self._reg == other._reg and self._width == other._width
        return False

    @property
    def bit_width(self) -> int:
        return self._width * 8

    @property
    def mask(self) -> int:
        bitrange = self._ida._ida_bitrange.bitrange_t(0, 64 * 8)
        name = self.name
        self._ida._ida_idp.get_reg_info(name, bitrange)
        if bitrange.empty():
            # documented as a special state meaning the value is all bits in the container
            return (1 << self.bit_width) - 1
        return int(bitrange.mask64())

    @property
    def base(self) -> IDARegister:
        bitrange = self._ida._ida_bitrange.bitrange_t(0, 64 * 8)
        name = self.name
        base_name = self._ida._ida_idp.get_reg_info(name, bitrange)
        if base_name and name != base_name:
            return self._ida.get_register(base_name)
        else:
            return self

    @property
    def name(self) -> str:
        return self._ida._ida_idp.get_reg_name(self._reg, self._width).lower()


class IDARegisterList(RegisterList):
    ...


class IDAARMPhrase(Phrase):
    """
    Defines an ARM phrase/displacement
    e.g.
        [R5],#4
        [R11,#-8]
    """

    def __init__(self, ida: IDAFlatAPI, insn_t: "ida_ua.insn_t", op_t: "ida_ua.op_t"):
        self._ida = ida
        self._insn_t = insn_t
        self._op_t = op_t
        self._width = ida.bit_size // 8

    @property
    def base(self) -> IDARegister:
        """
        The base register
        """
        return IDARegister(self._ida, self._op_t.reg, self._width)

    @property
    def index(self) -> Optional[IDARegister]:
        """
        The index register
        """
        # Index register is not a thing for ARM.
        return None

    @property
    def scale(self) -> int:
        """
        The scaling factor for the index.
        """
        return 1

    @property
    def offset(self) -> Union[IDARegister, int]:
        """
        The offset or displacement.
        This could be a register or immediate.
        e.g.
            [R1, R2] -> R2
            [R1, #1] -> 1

        NOTE: For shift information, please access the Operand.shift attribute.
            [R1, R2,LSL #3] -> R2
        """
        # [R1, R2]
        if self._op_t.type == self._ida._ida_ua.o_phrase:
            second_reg = self._ida._ida_arm.secreg(self._op_t)  # pulling the R2
            return IDARegister(self._ida, second_reg, self._width)

        # [R1, #1]
        else:
            offset = self._op_t.addr
            # Convert to signed number.
            if offset >> (self._ida.bit_size - 1):  # Is the hi-bit set?
                offset -= (1 << self._ida.bit_size)
            return offset


class IDAx86Phrase(Phrase):
    """
    Defines a x86 phrase/displacement.
    e.g.
        [ebp-eax*2+0x100]
        [ebp+4]
        fs:[eax]
    """

    def __init__(self, ida: IDAFlatAPI, insn_t: "ida_ua.insn_t", op_t: "ida_ua.op_t"):
        self._ida = ida
        self._insn_t = insn_t
        self._op_t = op_t
        self._width = ida.bit_size // 8

    @property
    def _segment_register(self) -> Optional[IDARegister]:
        """Obtains the fs/gs register if used."""
        # segpref holds the segment register for the x86 processor.
        if segpref := self._insn_t.segpref:
            # FIXME: Bit of a hack because I couldn't figure out how to properly determine which operand
            #   the segment register belonged to.
            reg_name = self._ida._ida_idp.get_reg_name(segpref, 2)
            text = self._ida._idc.print_operand(self._insn_t.ea, self._op_t.n)
            if reg_name in text:
                return IDARegister(self._ida, segpref, 2)

    def _x86_base_reg(self) -> Optional[IDARegister]:
        base_reg = self._ida._ida_intel.x86_base_reg(self._insn_t, self._op_t)
        if base_reg == -1:
            return None
        return IDARegister(self._ida, base_reg, self._width)

    @property
    def base(self) -> Optional[IDARegister]:
        """
        The base register.
        e.g.
            [ebp+ecx*2+var_8] -> ebp
            fs:[eax] -> fs
        """
        if seg_reg := self._segment_register:
            return seg_reg
        return self._x86_base_reg()

    @property
    def index(self) -> Optional[IDARegister]:
        """
        The index register
        e.g.
            [ebp+ecx*2+var_8] -> ecx
        """
        index_reg = self._ida._ida_intel.x86_index_reg(self._insn_t, self._op_t)
        if index_reg == -1:
            return None
        return IDARegister(self._ida, index_reg, self._width)

    @property
    def scale(self) -> int:
        """
        The scaling factor for the index.
        NOTE: This should default to 0 if index * scale is not supported in the processor. (ARM)

        e.g.
            [ebp+ecx*2+var_8] -> 2
        """
        return 1 << self._ida._ida_intel.sib_scale(self._op_t)

    @property
    def offset(self) -> Union[int, IDARegister]:
        """
        The offset or displacement.
        This could be a register or immediate.

        e.g.
            [ebp+ecx*2+var_8] -> var_8 -> 8
            fs:[eax] -> eax
        """
        # If we have a segment register, the traditional base register is actually the offset.
        if self._segment_register:
            return self._x86_base_reg()

        offset = self._op_t.addr
        # Convert to signed number.
        if offset >> (self._ida.bit_size - 1):  # Is the hi-bit set?
            offset -= (1 << self._ida.bit_size)
        return offset




from __future__ import annotations
import abc
from typing import List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from dragodis.interface.data_type import DataType
    from dragodis.interface.function_argument_location import ArgumentLocation


class FunctionSignature(metaclass=abc.ABCMeta):
    """
    Interface for a function signature.
    """

    def __str__(self) -> str:
        return self.declaration

    def __repr__(self) -> str:
        return f"<FunctionSignature {self.return_type} {self.calling_convention} {self.name}({', '.join(map(repr, self.parameters))})>"

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        The demangled name of the function.
        """

    @property
    @abc.abstractmethod
    def declaration(self) -> str:
        """
        The full function declaration.
        """

    @property
    @abc.abstractmethod
    def calling_convention(self) -> str:
        """
        The calling convention used in the function signature.
        """

    @calling_convention.setter
    @abc.abstractmethod
    def calling_convention(self, name: str):
        """
        Sets the calling convention for the function signature.
        """

    @property
    @abc.abstractmethod
    def return_type(self) -> DataType:
        """
        The return type of the function.
        """

    @return_type.setter
    @abc.abstractmethod
    def return_type(self, data_type: Union[DataType, str]):
        """
        Sets the return type of the function with given data type.
        """

    @property
    @abc.abstractmethod
    def parameters(self) -> List[FunctionParameter]:
        """
        List of parameters in the function signature.
        """

    @property
    def parameter_types(self) -> List[str]:
        """
        Convenience function for getting the parameter data types.
        This is useful when modifying the parameters.
        """
        return [param.data_type.name for param in self.parameters]

    @abc.abstractmethod
    def replace_parameters(self, data_types: List[str]):
        """
        Replaces the parameters of the signature with the ones given.

        :param data_types: List of strings representing the data types for new parameter.
        """

    @abc.abstractmethod
    def add_parameter(self, data_type: str) -> FunctionParameter:
        """
        Adds a parameter to the end of the signature set to given data type.

        :param data_type: Data type of new parameter.
        :returns: New parameter created.
        """

    @abc.abstractmethod
    def remove_parameter(self, ordinal: int):
        """
        Removes the parameter at the given ordinal.

        :param ordinal: Index of parameter
            (if negative, ordinal is relative to number of parameters)
        :raises NotExistError: If parameter at given ordinal doesn't exist.
        """

    @abc.abstractmethod
    def insert_parameter(self, ordinal: int, data_type: str) -> FunctionParameter:
        """
        Inserts a parameter at the given index set with the given data type.

        :param ordinal: Index of parameter
            (if negative, ordinal is relative to number of parameters)
        :param data_type: Data type of new parameter.
        :raises ValueError: If ordinal is out of range.
        :returns: New parameter created.
        """


class FunctionParameter(metaclass=abc.ABCMeta):
    """
    Interface for a parameter from a FunctionSignature
    """

    def __init__(self, signature: FunctionSignature):
        self.signature = signature

    def __str__(self) -> str:
        return f"{self.location}: {self.declaration}"

    def __repr__(self) -> str:
        return f"<FunctionParameter {self.ordinal}: {self}>"

    # TODO: Provide ability to change name?
    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Name of the parameter
        """

    @name.setter
    @abc.abstractmethod
    def name(self, new_name: str):
        """
        Sets name for the parameter.
        """

    @property
    @abc.abstractmethod
    def ordinal(self) -> int:
        """
        Index position within the signature.
        """

    @property
    def size(self) -> int:
        """
        The size of the parameter based on data type.
        """
        return self.data_type.size

    @property
    @abc.abstractmethod
    def data_type(self) -> DataType:
        """
        Data type for the parameter.
        """

    @data_type.setter
    @abc.abstractmethod
    def data_type(self, new_type: Union[str, DataType]):
        """
        Sets the data type for the parameter.

        :param new_type: Either a string or DataType object.
        """

    @property
    @abc.abstractmethod
    def declaration(self) -> str:
        """
        The parameter declaration as seen in the function signature.
        """

    @property
    @abc.abstractmethod
    def location(self) -> ArgumentLocation:
        """
        The location where the defined argument would be stored.
        e.g. stack, register, etc.
        """




from __future__ import annotations
import abc
import functools
from typing import Iterable, TYPE_CHECKING, Optional, Set

from dragodis.exceptions import NotExistError

if TYPE_CHECKING:
    from dragodis.interface import FlowType, Line, FlatAPI


@functools.total_ordering
class BasicBlock(metaclass=abc.ABCMeta):

    def __init__(self, api: FlatAPI):
        self._api = api

    def __str__(self) -> str:
        return f"block[0x{self.start:08x} --> 0x{self.end:08x}]"

    def __repr__(self) -> str:
        return f"<BasicBlock 0x{self.start:08x} --> 0x{self.end:08x}>"

    def __hash__(self):
        return hash(self.start)

    def __eq__(self, other) -> bool:
        """
        Checks equality of Basic Blocks.
        Defaults to checking that the start addresses are the same.
        """
        return isinstance(other, self.__class__) and self.start == other.start

    def __lt__(self, other):
        return self.start < other.start

    def __contains__(self, addr: int) -> bool:
        """
        Returns whether the given address is within the block.
        Defaults to checking if address is in-between start and end.
        """
        return self.start <= addr < self.end

    def __len__(self) -> int:
        """
        Returns the number of lines within the blocks.
        Defaults to checking length of self.lines
        """
        return len(list(self.lines()))

    @property
    @abc.abstractmethod
    def start(self) -> int:
        """
        The start address for the block.
        """

    @property
    @abc.abstractmethod
    def end(self) -> int:
        """
        The end address for the block.
        The end address is *not* part of the block.
        """

    @property
    @abc.abstractmethod
    def flow_type(self) -> FlowType:
        """
        The type of code flow for how things exit this block.
        """

    @property
    @abc.abstractmethod
    def flowchart(self) -> "Flowchart":
        """
        The Flowchart this block is contained in.
        """

    @property
    @abc.abstractmethod
    def blocks_to(self) -> Iterable["BasicBlock"]:
        """
        Iterates BasicBlocks that flow to this block.
        """

    @property
    @abc.abstractmethod
    def blocks_from(self) -> Iterable["BasicBlock"]:
        """
        Iterates BasicBlocks that flow from this block.
        """

    # TODO: Rename to instructions?
    def lines(self, start=None, reverse=False) -> Iterable[Line]:
        """
        Iterates Lines within the block.

        :param start: Start address (defaults to start or end address)
        :param reverse: Direction to iterate

        :yields: Line objects

        :raises NotExistError: If given start address is not in block
        """
        if start is not None and start not in self:
            raise NotExistError(f"Start address {hex(start)} is not in block.")

        if reverse:
            start = start or (self.end - 1)
            end = self.start - 1
        else:
            start = start or self.start
            end = self.end

        yield from self._api.lines(start=start, end=end, reverse=reverse)

    @property
    def ancestors(self) -> Set["BasicBlock"]:
        """
        Returns a set of ancestor blocks for the given block.

        :returns: Set of ancestor blocks.
        """
        ancestors = set()

        to_process = [self]
        while to_process:
            block = to_process.pop()
            for parent in block.blocks_to:
                if parent not in ancestors and parent != self:
                    ancestors.add(parent)
                    to_process.append(parent)

        return ancestors


class Flowchart(metaclass=abc.ABCMeta):

    def __len__(self) -> int:
        """
        Returns the number of basic blocks within the flowchart.
        Defaults to checking length of self.blocks
        """
        return sum(1 for _ in self.blocks)

    def __contains__(self, block: BasicBlock):
        """
        Tests if given block is in the flowchart.
        Overwrite for optimized implementation.
        """
        for _block in self.blocks:
            if _block == block:
                return True
        return False

    def __eq__(self, other):
        """
        Tests equality of Flowchart.
        Defaults to checking equality of all BasicBlocks.
        """
        if not isinstance(other, self.__class__):
            return False
        return all(b1 == b2 for b1, b2 in zip(self.blocks, other.blocks))

    def __str__(self) -> str:
        # Get first block to indicate address.
        start = self.start
        start = f"0x{start:08x}" if start is not None else "Empty"
        return f"flowchart[{start}]"

    def __repr__(self):
        blocks = "\t\n".join(map(repr, self.blocks))
        return f"<Flowchart \n\t{blocks}\n>"

    # TODO: This function should enforce that the order of the blocks is by address?
    #   I.E. Whatever the order IDA does.
    #   This will avoid the need for having a "start_block".
    @property
    @abc.abstractmethod
    def blocks(self) -> Iterable[BasicBlock]:
        """
        Iterates BasicBlocks within the flowchart.
        Blocks should be yielded in order of address. (ie. the order you see them in list view)
        """

    def _traverse(self, start_ea=None, dfs=False):
        """
        Blind traversal of the graph.

        :param int start_ea: EA within a block from which to start traversing
        :param bool dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.

        :yield BasicBlock: function block object
        """
        # Set our flag to True if start_ea is none so we yield all blocks, else wait till we find the requested block
        block_found = start_ea is None

        # Pull starting block.
        for block in self.blocks:
            non_visited = [block]
            break
        else:
            return

        visited = set()
        while non_visited:
            cur_block = non_visited.pop(0)
            if hash(cur_block) in visited:
                continue

            visited.add(hash(cur_block))
            succs = sorted(cur_block.blocks_from)
            if dfs:
                # [0:0] allows us to extend to the front
                non_visited[0:0] = succs
            else:
                non_visited.extend(succs)

            if not block_found:
                block_found = start_ea in cur_block

            if block_found:
                yield cur_block

    def _traverse_reverse(self, start=None, dfs=False) -> Iterable[BasicBlock]:
        """
        Perform a reverse traversal of the graph in depth-first/breadth-first manner where given a start node, traverse 1 complete
        path to the root node before following additional paths.

        :param start: Address within a block from which to start traversing
        :param dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.
        """
        if start:
            non_visited = [self.get_block(start)]
        else:
            non_visited = sorted(self.blocks, key=lambda bb: bb.start)[-1:]

        visited = set()
        while non_visited:
            cur_block = non_visited.pop(0)
            if hash(cur_block) in visited:
                continue

            visited.add(hash(cur_block))

            preds = sorted(cur_block.blocks_to, reverse=True)
            # For now, only consider predicates that are before the current block.
            # This helps to prevent cyclic loops.
            preds = [pred for pred in preds if pred < cur_block]
            if dfs:
                non_visited[0:0] = preds
            else:
                non_visited.extend(preds)

            yield cur_block

    def traverse(self, start=None, reverse=False, dfs=False) -> Iterable[BasicBlock]:
        """
        Iterates over basic blocks within the same order as code flow.

        :param start: optional start address to start iterating from.
        :param reverse: iterate in reverse.
        :param dfs: If true, traversal will be depth-first. If false, traversal will be breadth-first.
        """
        if reverse:
            yield from self._traverse_reverse(start, dfs=dfs)
        else:
            yield from self._traverse(start, dfs=dfs)

    def lines(self, start=None, reverse=False, dfs=False):
        """
        """
        _first_block = True
        for block in self.traverse(start, reverse=reverse, dfs=dfs):
            if start and _first_block:
                yield from block.lines(start, reverse=reverse)
            else:
                yield from block.lines(reverse=reverse)
            _first_block = False

    def get_block(self, addr: int) -> BasicBlock:
        """
        Gets BasicBlock containing given address.
        Defaults to iterating each block and checking if address is contained within.

        :raises NotExistError: If block doesn't exist with in the flowchart for the given address.
        """
        for block in self.blocks:
            if addr in block:
                return block
        raise NotExistError(f"Unable to find block containing address 0x{addr:08x} within flowchart.")

    @property
    def start(self) -> Optional[int]:
        """
        The first address or the first block, which should be the entry point
        or None if flowchart contains no blocks.
        (Override this function, if that is not the case.)
        """
        for block in self.blocks:
            return block.start



from __future__ import annotations
import logging
from typing import Optional, TYPE_CHECKING

from dragodis.exceptions import *
from dragodis.ida.flowchart import IDAFlowchart
from dragodis.ida.stack import IDAStackFrame
from dragodis.interface import Function, CommentType

if TYPE_CHECKING:
    import ida_funcs
    from dragodis.ida.flat import IDAFlatAPI


logger = logging.getLogger(__name__)


# TODO: Cache this function to return same object for same address.
#   (However, different addresses can be the same function!!!)
class IDAFunction(Function):

    def __init__(self, ida: IDAFlatAPI, func_t: "ida_funcs.func_t"):
        super().__init__(ida)
        self._ida = ida
        self._func_t = func_t
        self._name = None

    def __contains__(self, addr: int) -> bool:
        # NOTE: We can't just test if it's between start and end because the
        # function might have fragmented function chunks.
        try:
            func = self._ida.get_function(addr)
        except Exception:  # TODO: catch appropriate exception as used above.
            return False
        # return func is self  # TODO: Should be able to do this when reuse caching is on
        return func.start == self.start

    @property
    def start(self) -> int:
        return self._func_t.start_ea

    @property
    def end(self) -> int:
        return self._func_t.end_ea

    @property
    def flowchart(self) -> IDAFlowchart:
        return IDAFlowchart(self._ida, self._ida._ida_gdl.FlowChart(self._func_t))

    def get_comment(self, comment_type=CommentType.anterior) -> Optional[str]:
        # If user asks for anterior or eol, assume they mean the same thing.
        if comment_type in (CommentType.anterior, CommentType.plate):
            return self._ida._ida_funcs.get_func_cmt(self._func_t, 0)
        elif comment_type == CommentType.repeatable:
            return self._ida._ida_funcs.get_func_cmt(self._func_t, 1)
        else:
            raise ValueError(f"Invalid comment type for function: {repr(comment_type)}")

    @property
    def name(self) -> str:
        if not self._name:
            self._name = self._ida._ida_funcs.get_func_name(self.start)
        return self._name

    @name.setter
    def name(self, new_name: Optional[str]):
        success = self._ida._ida_name.set_name(
            self.start, new_name or "", self._ida._ida_name.SN_NOCHECK)
        if not success:
            raise ValueError(f"Failed to set function name at {hex(self.start)} with {new_name}")
        self._name = None  # clear cache

    def set_comment(self, comment: str, comment_type=CommentType.anterior):
        # IDA takes an empty string to clear comments
        if not comment:
            comment = ""

        # If user asks for anterior or eol, assume they mean the same thing.
        if comment_type in (CommentType.anterior, CommentType.plate):
            self._ida._ida_funcs.set_func_cmt(self._func_t, comment, 0)
        elif comment_type == CommentType.repeatable:
            self._ida._ida_funcs.set_func_cmt(self._func_t, comment, 1)
        else:
            raise ValueError(f"Invalid comment type for function: {repr(comment_type)}")

    @property
    def source_code(self) -> Optional[str]:
        decompiled_code = self._ida._ida_helpers.decompiled_code(self.start)
        if decompiled_code:
            return str(decompiled_code)

    @property
    def stack_frame(self) -> IDAStackFrame:
        return IDAStackFrame(self._ida, self._ida._ida_frame.get_frame(self._func_t))

    @property
    def is_library(self) -> bool:
        return bool(self._func_t.flags & self._ida._ida_funcs.FUNC_LIB)



from __future__ import annotations
from typing import TYPE_CHECKING, Union, Optional

from dragodis.interface.operand_value import (
    OperandValue, Immediate, MemoryReference, Register, RegisterList,
    Phrase,
)


if TYPE_CHECKING:
    import ghidra


class GhidraImmediate(Immediate):
    ...


class GhidraMemoryReference(MemoryReference):
    ...


class GhidraRegister(Register):

    def __init__(self, register: "ghidra.program.model.lang.Register"):
        self._register = register

    def __eq__(self, register: "GhidraRegister"):
        if isinstance(register, GhidraRegister):
            return self._register.equals(register._register)
        return False

    @property
    def bit_width(self) -> int:
        return int(self._register.getBitLength())

    @property
    def mask(self) -> int:
        mask = bytes(self._register.getBaseMask())
        return int.from_bytes(mask, byteorder='big', signed=False)

    @property
    def base(self) -> GhidraRegister:
        return GhidraRegister(self._register.getBaseRegister())

    @property
    def name(self) -> str:
        return str(self._register.getName()).lower()


class GhidraRegisterList(RegisterList):
    ...


class GhidraPhrase(Phrase):
    """
    Defines general "phrase" operand in Ghidra.
    """

    def __init__(
            self,
            base: Optional[GhidraRegister] = None,
            index: Optional[GhidraRegister] = None,
            scale: int = 1,
            offset: Union[GhidraRegister, int] = 0,
    ):
        """
        Initializes phrase with given base, index, scale, and offset
        (from GhidraOperand)
        """
        self._base = base
        self._index = index
        # Wrapping in int() to remove Java nonsense.
        self._scale = int(scale)
        if not isinstance(offset, GhidraRegister):
            offset = int(offset)
        self._offset = offset

    @property
    def base(self) -> Optional[GhidraRegister]:
        return self._base

    @property
    def index(self) -> Optional[GhidraRegister]:
        return self._index

    @property
    def scale(self) -> int:
        return self._scale

    @property
    def offset(self) -> Union[GhidraRegister, int]:
        return self._offset




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable

from dragodis.ida.line import IDALine
from dragodis.ida.memory import IDAMemory
from dragodis.interface import Segment, SegmentPermission

if TYPE_CHECKING:
    import ida_segment
    from dragodis.ida.flat import IDAFlatAPI


class IDASegment(Segment):

    def __init__(self, ida: IDAFlatAPI, segment_t: "ida_segment.segment_t"):
        self._ida = ida
        self._segment_t = segment_t
        self.__real_end = None  # caching for end address.

    @property
    def name(self) -> str:
        return self._ida._ida_segment.get_segm_name(self._segment_t)

    @property
    def start(self) -> int:
        return self._segment_t.start_ea

    @property
    def _real_end(self) -> int:
        """
        Returns the address for the real end of the segment without the padding.
        """
        if not self.initialized:
            return self.start
        if self.__real_end is None:
            # Exclude any overlay of uninitialized bytes from the segment.
            # This is to better match Ghidra's approach of separating uninitalized sections
            # into their own block.
            ida_bytes = self._ida._ida_bytes
            end = self._segment_t.end_ea
            if not ida_bytes.is_loaded(end - 1):
                # Find first initialized item, then move down to find first uninitialized byte from that.
                end = ida_bytes.prev_inited(end - 1, self._segment_t.start_ea)
                while ida_bytes.is_loaded(end):
                    end += 1
            self.__real_end = end
        return self.__real_end

    @property
    def end(self) -> int:
        return self._segment_t.end_ea

    @property
    def initialized(self) -> bool:
        # If first address isn't loaded, then no bytes are.
        return self._ida._ida_bytes.is_loaded(self._segment_t.start_ea)

    @property
    def bit_size(self) -> int:
        return self._segment_t.abits()

    @property
    def permissions(self) -> SegmentPermission:
        perm = self._segment_t.perm
        ret = SegmentPermission(0)
        if perm & self._ida._ida_segment.SEGPERM_EXEC:
            ret |= SegmentPermission.execute
        if perm & self._ida._ida_segment.SEGPERM_WRITE:
            ret |= SegmentPermission.write
        if perm & self._ida._ida_segment.SEGPERM_READ:
            ret |= SegmentPermission.read
        return ret

    @property
    def lines(self) -> Iterable[IDALine]:
        if self.initialized:
            yield from self._ida.lines(self.start, self._real_end)

    def open(self) -> IDAMemory:
        if not self.initialized:
            # Empty memory
            return self._ida.open_memory(self.start, self.start)
        return self._ida.open_memory(self.start, self._real_end)




from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Iterable

from dragodis.ghidra.flowchart import GhidraFlowchart
from dragodis.ghidra.function_signature import GhidraFunctionSignature
from dragodis.ghidra.instruction import GhidraInstruction
from dragodis.ghidra.reference import GhidraReference
from dragodis.ghidra.stack import GhidraStackFrame
from dragodis.interface import Function, CommentType

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraFunction(Function):

    def __init__(self, ghidra: GhidraFlatAPI, function: "ghidra.program.model.listing.Function"):
        super().__init__(ghidra)
        self._ghidra = ghidra
        self._function = function

    def __contains__(self, addr: int) -> bool:
        return self._body.contains(self._ghidra._to_addr(addr))

    @property
    def _body(self) -> "ghidra.program.model.address.AddressSetView":
        return self._function.getBody()

    @property
    def start(self) -> int:
        return self._function.getEntryPoint().getOffset()

    @property
    def end(self) -> int:
        # NOTE: Calling next() to get the address AFTER the last instruction.
        #   This is to replicate the non-inclusionary characteristics of IDA.
        return self._body.getMaxAddress().next().getOffset()

    @property
    def flowchart(self) -> GhidraFlowchart:
        return GhidraFlowchart(self._ghidra, self._body)

    def get_comment(self, comment_type=CommentType.plate) -> Optional[str]:
        if comment_type in (CommentType.anterior, CommentType.plate):
            return self._function.getComment()
        elif comment_type == CommentType.repeatable:
            return self._function.getRepeatableComment()
        else:
            raise ValueError(f"Invalid comment type for function: {repr(comment_type)}")

    @property
    def name(self) -> str:
        return self._function.getName()

    @name.setter
    def name(self, value: Optional[str]):
        from ghidra.program.model.symbol import SourceType
        self._function.setName(value, SourceType.USER_DEFINED)

    def set_comment(self, comment: Optional[str], comment_type=CommentType.plate):
        if comment_type in (CommentType.anterior, CommentType.plate):
            self._function.setComment(comment)
        elif comment_type == CommentType.repeatable:
            self._function.setRepeatableComment(comment)
        else:
            raise ValueError(f"Invalid comment type for function: {repr(comment_type)}")

    @property
    def source_code(self) -> Optional[str]:
        code = self._ghidra._decomp_api.decompile(self._function)
        # Remove carriage returns, leaving only the newlines.
        code = code.replace("\r", "")
        return code

    @property
    def stack_frame(self) -> GhidraStackFrame:
        return GhidraStackFrame(self._ghidra, self._function.getStackFrame())

    @property
    def signature(self) -> GhidraFunctionSignature:
        return GhidraFunctionSignature(self._ghidra, self._function)

    @property
    def is_library(self) -> bool:
        return self._function in self._ghidra._static_functions




from __future__ import annotations
from typing import Optional, TYPE_CHECKING, Union, Tuple

from dragodis.exceptions import NotExistError
from dragodis.ida.stack import IDAStackVariable, IDAStackFrame
from dragodis.ida.variable import IDAVariable
from dragodis.interface import Phrase
from dragodis.interface.operand import Operand, ARMOperand, x86Operand, OperandType
from dragodis.ida.operand_value import (
    OperandValue, IDARegister, IDAImmediate, IDAMemoryReference,
    IDAARMPhrase, IDAx86Phrase, IDARegisterList,
)
from dragodis.interface.types import ARMShiftType

if TYPE_CHECKING:
    import ida_ua
    from dragodis.ida.flat import IDAFlatAPI
    from dragodis.ida.instruction import IDAInstruction


# TODO: Cache this operand to return same object for same address.
class IDAOperand(Operand):

    _type_map = {
        0: OperandType.void,          # o_void
        1: OperandType.register,      # o_reg
        2: OperandType.memory,        # o_mem
        3: OperandType.phrase,        # o_phrase
        4: OperandType.phrase,        # o_displ
        5: OperandType.immediate,     # o_imm
        6: OperandType.code,          # o_far
        7: OperandType.code,          # o_near
    }

    def __init__(self, instruction: IDAInstruction, ida: IDAFlatAPI, addr: int, index: int, op_t: "ida_ua.op_t"):
        super().__init__(instruction)
        self._ida = ida
        self._addr = addr
        self._index = index
        self._op_t = op_t

    @property
    def address(self) -> int:
        return self._addr

    @property
    def index(self) -> int:
        return self._index

    @property
    def text(self) -> str:
        # Get operand text and then remove the color tags.
        # (Doing same thing as idc.print_operand())
        text = self._ida._ida_ua.print_operand(self.address, self.index)
        if text:
            return self._ida._ida_lines.tag_remove(text)

        # No text indicates an implied operand, which *should* be a register but could be a phrase
        _value = self.value
        if isinstance(_value, IDARegister):
            return _value.name

        if isinstance(_value, IDAx86Phrase):
            return _value.base

        raise AssertionError(f"Expected operand to be register or phrase, got: {_value!r}")

    @property
    def type(self) -> OperandType:
        # NOTE: Getting operand type using get_operand_type() instead of _op_t.type because
        #   there are strange issues where `_op_t.type` will sometimes give us something completely wrong.
        op_type = self._ida._idc.get_operand_type(self._addr, self._index)
        try:
            return self._type_map[op_type]
        except KeyError:
            raise RuntimeError(f"Unexpected operand type: {op_type}")

    @property
    def value(self) -> OperandValue:
        operand_type = self.type
        # TODO: Should we be recording both .value and .addr in the MemoryReference object?
        if operand_type in (OperandType.memory, OperandType.code):  # o_mem or o_near/o_far
            return IDAMemoryReference(self._op_t.addr)
        elif operand_type == OperandType.register:  # o_reg
            return IDARegister(self._ida, self._op_t.reg, self.width)
        elif operand_type == OperandType.immediate:  # o_imm
            # value = self._op_t.value
            value = self._ida._idc.get_operand_value(self._addr, self._index)
            # Need to mask off the value based on width since IDA will sometimes
            # include ff bytes in the front causing the value to be incorrect.
            value &= (1 << (8 * self.width)) - 1
            return IDAImmediate(value)

        # Architecture specific operands types like phrase should be handled by the
        # appropriate subclass.
        raise ValueError(f"Invalid operand type: {operand_type!r} @ {hex(self.address)}:{self.index}")

    @property
    def width(self) -> int:
        return self._ida._ida_ua.get_dtype_size(self._op_t.dtype)

    @property
    def variable(self) -> Optional[IDAVariable]:
        value = self.value
        if isinstance(value, Phrase):
            offset = value.offset
            if isinstance(offset, int):
                stack_var = self._ida._ida_frame.get_stkvar(self.instruction._insn_t, self._op_t, value.offset)
                if stack_var:
                    member, _ = stack_var
                    func_t = self._ida._ida_funcs.get_func(self.address)
                    frame = self._ida._ida_frame.get_frame(func_t)
                    frame = IDAStackFrame(self._ida, frame)
                    return IDAStackVariable(self._ida, frame, member)

        elif isinstance(value, int):  # memory or immediate
            try:
                # Cast to int() to strip off custom classes.
                return self._ida.get_variable(int(value))
            except NotExistError:
                pass


class IDAARMOperand(IDAOperand, ARMOperand):

    # Append ARM specific operand types.
    _type_map = {
        **IDAOperand._type_map,
        # We treat shifted registers just as regular register types, but with the .shift attribute available.
        8: OperandType.register,       # ida_arm.o_shreg
        9: OperandType.register_list,  # ida_arm.o_reglist
    }

    @property
    def shift(self) -> Tuple[ARMShiftType, Union[int, IDARegister]]:
        if self._op_t.type == self._ida._ida_ua.o_phrase:
            # For a phrase, shift count is in op.value
            shift_count = self._op_t.value
            return ARMShiftType(self._op_t.specflag2), shift_count

        elif self._op_t.type == self._ida._ida_arm.o_shreg:
            # Shift can be an immediate or another register.
            # I believe we determine which one it is based on whether op.value is zero.
            shift_count = self._op_t.value
            if not shift_count:
                shift_count = IDARegister(self._ida, self._ida._ida_arm.secreg(self._op), self.width)
            return ARMShiftType(self._op_t.specflag2), shift_count

        return ARMShiftType.LSL, 0  # not shifted

    @property
    def value(self) -> OperandValue:
        # Get value for ARM specific types.
        operand_type = self.type

        if operand_type == OperandType.register_list:
            # Register numbers are stored in a bitmap in specval.
            reg_bitmap = self._op_t.specval
            width = self.width
            return IDARegisterList([
                IDARegister(self._ida, reg, width)
                for reg in range(16) if reg_bitmap & (1 << reg)
            ])

        if operand_type == OperandType.phrase:  # o_phrase/o_displ
            return IDAARMPhrase(self._ida, self.instruction._insn_t, self._op_t)

        return super().value


class IDAx86Operand(IDAOperand, x86Operand):

    @property
    def type(self) -> OperandType:
        # For x86, there is a weird corner case, where we could have something like
        # dword_40DC20[eax*4], which really is closer to a phrase, without
        # the base:  [0 + eax*4 + dword_40DC20]
        type = super().type
        if type == OperandType.memory:
            try:
                # Check if we have an index register to know if if this should
                # be a phrase.
                index_reg = self._ida._ida_intel.x86_index_reg(self.instruction._insn_t, self._op_t)
                if index_reg != -1:  # RegNo.R_none
                    return OperandType.phrase
            except ValueError:
                pass
        return type

    @property
    def value(self) -> OperandValue:
        # Get value for x86 specific types.
        operand_type = self.type

        if operand_type == OperandType.phrase:  # o_phrase/o_displ
            return IDAx86Phrase(self._ida, self.instruction._insn_t, self._op_t)

        return super().value



'''
Program Variable
'''

from abc import abstractmethod, ABCMeta
from dragodis.interface.data_type import DataType


class Variable(metaclass=ABCMeta):
    """Function Local/Global Variable"""

    def __str__(self) -> str:
        return f"{self.data_type} {self.name}"

    def __repr__(self) -> str:
        return f"<Variable {self} - size={self.size}>"

    @property
    @abstractmethod
    def name(self) -> str:
        """The variable's name"""

    @name.setter
    @abstractmethod
    def name(self, new_name: str):
        """Sets the variable's name"""

    @property
    @abstractmethod
    def size(self) -> int:
        """The variable's size"""

    @property
    @abstractmethod
    def data_type(self) -> DataType:
        """The variable's type"""

    # @data_type.setter
    # @abstractmethod
    # def data_type(self, data_type: DataType):
    #     """Sets the variable's data type"""


class GlobalVariable(Variable, metaclass=ABCMeta):
    """Global variable usually defined in the .data section."""

    def __str__(self) -> str:
        return f"0x{self.address:08x}: {super().__str__()}"

    def __repr__(self) -> str:
        return f"<GlobalVariable {self} - size={self.size}>"

    @property
    @abstractmethod
    def address(self) -> int:
        """
        Address where the variable data is mapped to.
        """


class StackVariable(Variable, metaclass=ABCMeta):
    """Function Stack/Local Variable"""

    def __str__(self) -> str:
        return f"stack[0x{self.stack_offset:x}]: {super().__str__()}"

    def __repr__(self) -> str:
        return f"<StackVariable {self} - size={self.size}>"

    @property
    @abstractmethod
    def stack_offset(self) -> int:
        """
        The stack offset if variable is in a stack frame.
        NOTE: We are considered the stack when initially inside the function.
        ie. stack offset of zero would contain the return address followed by
        the arguments if incrementing.
        """




from __future__ import annotations
import abc
from enum import IntEnum, auto
from typing import TYPE_CHECKING, List, Iterable, Type, Optional

import capstone

from dragodis.interface.types import FlowType
from dragodis.interface.operand import Operand, ARMOperand, x86Operand

if TYPE_CHECKING:
    from dragodis.interface import Reference, FlatAPI, Line, Variable


class Instruction(metaclass=abc.ABCMeta):
    """
    Instruction objects represent the actual assembly code of a line of code.
    """
    # Subclasses should update these to change which class gets used
    # depending on detected processor.
    _Operand: Type[Operand] = Operand
    _ARMInstruction: Type[Instruction] = None
    _x86Instruction: Type[Instruction] = None

    def __new__(cls, api: FlatAPI, *args, **kwargs):
        """
        During instantiation, the Instruction constructor will dynamically
        choose the correct class based on detected processor.
        """
        if api.processor_name.startswith("ARM"):
            klass = cls._ARMInstruction
        elif api.processor_name.startswith("x86"):
            klass = cls._x86Instruction
        else:
            klass = cls

        self = super().__new__(klass)
        self.__init__(api, *args, **kwargs)
        return self

    def __init__(self, api: FlatAPI):
        self._api = api

    def __repr__(self):
        return (
            f"<Instruction 0x{self.address:08x}"
            f" - {self.text}"
            f">"
        )

    def __str__(self):
        """Displays instruction text"""
        return self.text

    @property
    def _capstone_insn(self) -> capstone.CsInsn:
        """
        Obtains the capstone instruction disassembly.
        This is used to subsidize some of the processor specific functions,
        in the case the backend disassembler can't handle getting the information.

        :raises RuntimeError: If we fail to disassemble the undelying bytes
            of the function. (Which shouldn't happen.)
        """
        # Default to using capstone to pull this information.
        for insn in self._api._capstone_dis.disasm(self.data, 0x1000):
            return insn
        raise RuntimeError(f"Failed to disassemble instruction at: 0x{self.start:08X}")

    @property
    @abc.abstractmethod
    def address(self) -> int:
        """
        The start address of the instruction.
        """

    @property
    def line(self) -> Line:
        """
        The defined line for this instruction.
        """
        return self._api.get_line(self.address)

    @property
    def data(self) -> bytes:
        """
        The bytes that make up this instruction.
        """
        return self.line.data

    @property
    @abc.abstractmethod
    def flow_type(self) -> FlowType:
        """
        The type of of code flow for the instruction.
        """

    @property
    def is_call(self) -> bool:
        """
        Is the instruction a call instruction
        """
        return self.flow_type == FlowType.call

    @property
    def is_jump(self) -> bool:
        """
        Is the instruction an indirect jump instruction
        """
        return self.flow_type in (FlowType.conditional_jump, FlowType.unconditional_jump)

    @property
    def is_return(self) -> bool:
        """
        Is the instruction a return instruction
        """
        return self.flow_type == FlowType.terminal

    @property
    @abc.abstractmethod
    def mnemonic(self) -> str:
        """
        The instruction's mnemonic.

        The output text is disassembler specific, but all implementations
        will likely be similar. However, each implementation should do its best
        to follow the official mnemonics as much as possible.
        Mnemonic should always be lowercase to keep consistency.

        :return: String representation of the mnemonic
        """

    @property
    @abc.abstractmethod
    def root_mnemonic(self) -> str:
        """
        The instruction's root mnemonic which doesn't include any pre/post-fix modifications.
        (e.g. the "bl" in "blne")
        """

    @property
    @abc.abstractmethod
    def operands(self) -> List[Operand]:
        """
        Returns a list of `Operand` objects contained within the instruction.
        Order of operands in list should match the index of the operands.
        """

    # TODO: The following were originally decided based on what was available to IDA.
    #   Look into adding more such as is_jump, is_fall_through, tc.

    # TODO: Should we have an InstructionType class?

    @property
    @abc.abstractmethod
    def text(self) -> str:
        """
        The disassembled assembly code for the instruction.

        WARNING: The output text is disassembler specific and therefore may
        vary in appearance based on disassembler used.

        :return: String representation of instruction
        """

    @property
    def references_from(self) -> Iterable[Reference]:
        """
        Iterates cross references from the specified address.

        :param int addr: Address to get references from
        :yield: `Reference` objects.
        """
        yield from self._api.references_from(self.address)

    @property
    def references_to(self) -> Iterable[Reference]:
        """
        Iterates cross references to the specified address.

        :param int addr: Address to get references to
        :yield: `Reference` objects.
        """
        yield from self._api.references_to(self.address)

    @property
    @abc.abstractmethod
    def stack_depth(self) -> int:
        """
        The depth of stack pointer at the beginning of the instruction.
        This is relative to the start of the function.
        (This is usually the number after the address in list view of the disassembler,
        but negative.)
        """

    @property
    @abc.abstractmethod
    def stack_delta(self) -> int:
        """
        The change in stack depth if the instruction was applied.
        """

    @property
    def variables(self) -> Iterable[Variable]:
        """
        Iterates the variables in the instruction.
        """
        for operand in self.operands:
            if variable := operand.variable:
                yield variable


class ARMConditionCode(IntEnum):
    INVALID = -1
    EQ = 0       # 0000 Z                        Equal
    NE = auto()  # 0001 !Z                       Not equal
    CS = auto()  # 0010 C                        Unsigned higher or same
    CC = auto()  # 0011 !C                       Unsigned lower
    MI = auto()  # 0100 N                        Negative
    PL = auto()  # 0101 !N                       Positive or Zero
    VS = auto()  # 0110 V                        Overflow
    VC = auto()  # 0111 !V                       No overflow
    HI = auto()  # 1000 C & !Z                   Unsigned higher
    LS = auto()  # 1001 !C | Z                   Unsigned lower or same
    GE = auto()  # 1010 (N & V) | (!N & !V)      Greater or equal
    LT = auto()  # 1011 (N & !V) | (!N & V)      Less than
    GT = auto()  # 1100 !Z & ((N & V)|(!N & !V)) Greater than
    LE = auto()  # 1101 Z | (N & !V) | (!N & V)  Less than or equal
    AL = auto()  # 1110 Always
    NV = auto()  # 1111 Never


class ARMInstruction(Instruction):
    _Operand = ARMOperand

    @property
    def update_flags(self) -> bool:
        """
        Whether the condition flags are updated on the result of the operation.
        (S postfix)
        """
        # Default to using capstone.
        return self._capstone_insn.update_flags

    @property
    def condition_code(self) -> ARMConditionCode:
        """
        Returns the condition flag.
        """
        # Default to using capstone.
        cc = self._capstone_insn.cc
        return ARMConditionCode(ARMConditionCode.INVALID + cc)

    @property
    def writeback(self) -> bool:
        """
        Whether the instruction contains an operand that updates itself
        before or after the value is obtained.
        """
        # Default to using capstone.
        # We count push and pop instructions as having a writeback because
        # some disassemblers interpret these as stmdb or ldmia opcodes with sp!
        insn = self._capstone_insn
        return insn.mnemonic.lower() in ("push", "pop") or insn.writeback

    @property
    def pre_indexed(self) -> bool:
        """
        Whether the instruction has a pre-indexed writeback.
        Ie, the register is updated before evaluation:

        .. code::

            [R1, 8]!
        """
        return self.writeback and "!" in self.text

    @property
    def post_indexed(self) -> bool:
        """
        Whether the instruction has a post-indexed writeback.
        Ie, the register is updated after evaluation:

        .. code::

            [R1], 8
        """
        return self.writeback and "!" not in self.text


class x86Instruction(Instruction):
    _Operand = x86Operand

    @property
    @abc.abstractmethod
    def rep(self) -> Optional[str]:
        """
        Rep prefix applied to instruction if provided.

        .. code::

            rep
            repne
        """


Instruction._ARMInstruction = ARMInstruction
Instruction._x86Instruction = x86Instruction




from __future__ import annotations

import pathlib

import atexit
import gc
import re
import logging
import os
import socket
import subprocess
import sys
import tempfile
import time
import typing
from typing import Callable
import uuid

import rpyc
from rpyc.core.stream import NamedPipeStream, SocketStream
from rpyc.utils import factory

from dragodis.exceptions import DragodisError, NotInstalledError
from dragodis.interface import BackendDisassembler
from dragodis.interface.types import ProcessorType
from dragodis.ida import ida_server, constants
from dragodis.constants import BACKEND_IDA
from dragodis import utils

# Used for typing to help Pycharm give us autocompletion.
# To setup, add %IDA_DIR%\python\3 directory into the interpreter path using Pycharm.
if typing.TYPE_CHECKING:
    import idaapi
    import idc
    import idautils
    import ida_auto
    import ida_bitrange
    import ida_bytes
    import ida_funcs
    import ida_gdl
    import ida_name
    import ida_xref
    import ida_ua
    import ida_idp
    import ida_ida
    import ida_lines
    import ida_nalt
    import ida_hexrays
    import ida_segment
    import ida_loader
    import ida_typeinf
    import ida_entry
    import ida_struct
    import ida_frame
    import ida_search
    from .sdk import ida_arm
    from .sdk import ida_intel
    from .sdk import ida_helpers

logger = logging.getLogger(__name__)


class _Cached:
    """
    Caches attributes for a given rpyc netref to avoid excess socket calls.
    """

    def __init__(self, netref: rpyc.BaseNetref):
        self._cache = {}
        self._netref = netref

    def __getattribute__(self, name):
        cache = super().__getattribute__("_cache")
        try:
            ret = cache[name]
            return ret
        except KeyError:
            netref = super().__getattribute__("_netref")
            ret = getattr(netref, name)
            cache[name] = ret
            return ret


class IDADisassembler(BackendDisassembler):
    """
    Backend IDA disassembler (interface)
    """
    name = BACKEND_IDA

    PROCESSOR_ARM = constants.PROCESSOR_ARM
    PROCESSOR_ARM64 = constants.PROCESSOR_ARM
    PROCESSOR_X86 = constants.PROCESSOR_METAPC
    PROCESSOR_X64 = constants.PROCESSOR_METAPC

    _BADADDR: int
    _idaapi: idaapi
    _idc: idc
    _idautils: idautils
    _ida_auto: ida_auto
    _ida_bitrange: ida_bitrange
    _ida_bytes: ida_bytes
    _ida_funcs: ida_funcs
    _ida_gdl: ida_gdl
    _ida_name: ida_name
    _ida_xref: ida_xref
    _ida_ua: ida_ua
    _ida_idp: ida_idp
    _ida_ida: ida_ida
    _ida_lines: ida_lines
    _ida_nalt: ida_nalt
    _ida_hexrays: ida_hexrays
    _ida_segment: ida_segment
    _ida_loader: ida_loader
    _ida_typeinf: ida_typeinf
    _ida_entry: ida_entry
    _ida_struct: ida_struct
    _ida_frame: ida_frame
    _ida_search: ida_search
    _ida_arm: ida_arm
    _ida_intel: ida_intel
    _ida_helpers: ida_helpers


class IDALocalDisassembler(IDADisassembler):
    """
    Backend used when we are natively in the IDA interpreter.
    """

    def __init__(self, input_path=None, **unused):
        import idc
        super().__init__(idc.get_input_file_path())

        # Input path is not required when inside IDA, but if provided
        # let's use it to validate we are looking at the right file.
        if input_path and pathlib.Path(input_path).resolve() != self.input_path:
            raise ValueError(
                f"Expected input path isn't the same as the file loaded in IDA: {input_path} != {self.input_path}")
        self.start()

    def start(self):
        import idc
        self._BADADDR = idc.BADADDR
        self._idc = idc
        import idaapi
        self._idaapi = idaapi
        import idautils
        self._idautils = idautils
        import ida_auto
        self._ida_auto = ida_auto
        import ida_bitrange
        self._ida_bitrange = ida_bitrange
        import ida_bytes
        self._ida_bytes = ida_bytes
        import ida_funcs
        self._ida_funcs = ida_funcs
        import ida_gdl
        self._ida_gdl = ida_gdl
        import ida_name
        self._ida_name = ida_name
        import ida_xref
        self._ida_xref = ida_xref
        import ida_ua
        self._ida_ua = ida_ua
        import ida_idp
        self._ida_idp = ida_idp
        import ida_ida
        self._ida_ida = ida_ida
        import ida_lines
        self._ida_lines = ida_lines
        import ida_nalt
        self._ida_nalt = ida_nalt
        import ida_hexrays
        self._ida_hexrays = ida_hexrays
        import ida_segment
        self._ida_segment = ida_segment
        import ida_loader
        self._ida_loader = ida_loader
        import ida_typeinf
        self._ida_typeinf = ida_typeinf
        import ida_entry
        self._ida_entry = ida_entry
        import ida_struct
        self._ida_struct = ida_struct
        import ida_frame
        self._ida_frame = ida_frame
        import ida_search
        self._ida_search = ida_search
        from .sdk import ida_arm
        self._ida_arm = ida_arm
        from .sdk import ida_intel
        self._ida_intel = ida_intel
        from .sdk import ida_helpers
        self._ida_helpers = ida_helpers

        idc.auto_wait()


class IDARemoteDisassembler(IDADisassembler):
    """
    Backend disassembler when we are remotely accessing IDA through rpyc.
    """

    _rpyc_config = {
        # Increasing timeout to give IDA more time to analyze larger files.
        "sync_request_timeout": 60
    }

    def __init__(self, input_path, is_64_bit=None, ida_path=None, timeout=None, processor=None, **unused):
        """
        Initializes IDA disassembler.

        :param input_path: Path of binary to process.
        :param is_64_bit: Whether input file is 64bit or 32bit.
            If left as None, this will be determined by analyzing the input file.
        :param ida_path: Path to IDA directory.
            This may also be set using the environment variable IDA_INSTALL_DIR
        :param timeout: Number of seconds to wait for remote results. (defaults to 60)
        :param processor: Processor type (defaults to auto detected)
            (https://hex-rays.com/products/ida/support/idadoc/618.shtml)
        """
        super().__init__(input_path, processor=processor)
        self._ida_path = ida_path or os.environ.get("IDA_INSTALL_DIR", os.environ.get("IDA_DIR"))
        if not self._ida_path:
            raise NotInstalledError(
                "Failed to get IDA install directory. "
                "Please provide it during instantiation or set the IDA_INSTALL_DIR environment variable."
            )
        self._script_path = ida_server.__file__
        if timeout is not None:
            self._rpyc_config = dict(self._rpyc_config)
            self._rpyc_config["sync_request_timeout"] = timeout

        # Determine if 64 bit.
        if is_64_bit is None:
            input_path = str(input_path)
            # First check if input file is a .idb or .i64
            if input_path.endswith(".i64"):
                is_64_bit = True
            elif input_path.endswith(".idb"):
                is_64_bit = False
            elif processor in (ProcessorType.ARM64, ProcessorType.x64):
                is_64_bit = True
            elif processor in (ProcessorType.ARM, ProcessorType.x86):
                is_64_bit = False
            else:
                is_64_bit = utils.is_64_bit(input_path)

        # Find ida executable within ida_dir.
        # TODO: Should we just always open with ida64?
        ida_exe_re = re.compile(r"idaq?64(\.exe)?$" if is_64_bit else r"idaq?(\.exe)?$")
        for filename in os.listdir(self._ida_path):
            if ida_exe_re.match(filename):
                self._ida_exe = os.path.abspath(os.path.join(self._ida_path, filename))
                break
        else:
            raise NotInstalledError(f"Unable to find ida executable within: {self._ida_path}")

        self._running = False

        self._socket_path = None
        self._process = None
        self._bridge = None
        self._root = None
        self._BADADDR = 0xFFFFFFFFFFFFFFFF if is_64_bit else 0xFFFFFFFF

        self._idaapi = None
        self._idc = None
        self._idautils = None
        self._ida_auto = None
        self._ida_bitrange = None
        self._ida_bytes = None
        self._ida_funcs = None
        self._ida_gdl = None
        self._ida_name = None
        self._ida_xref = None
        self._ida_ua = None
        self._ida_idp = None
        self._ida_ida = None
        self._ida_lines = None
        self._ida_nalt = None
        self._ida_hexrays = None
        self._ida_segment = None
        self._ida_loader = None
        self._ida_typeinf = None
        self._ida_entry = None
        self._ida_struct = None
        self._ida_frame = None
        self._ida_search = None
        self._ida_arm = None
        self._ida_intel = None
        self._ida_helpers = None

        # Used for typing to help Pycharm give us autocompletion.
        # This code shouldn't actually be run.
        # noinspection PyUnreachableCode
        if False:
            self._idaapi = idaapi
            self._idc = idc
            self._idautils = idautils
            self._ida_auto = ida_auto
            self._ida_bitrange = ida_bitrange
            self._ida_bytes = ida_bytes
            self._ida_funcs = ida_funcs
            self._ida_gdl = ida_gdl
            self._ida_name = ida_name
            self._ida_xref = ida_xref
            self._ida_ua = ida_ua
            self._ida_idp = ida_idp
            self._ida_ida = ida_ida
            self._ida_lines = ida_lines
            self._ida_nalt = ida_nalt
            self._ida_hexrays = ida_hexrays
            self._ida_segment = ida_segment
            self._ida_loader = ida_loader
            self._ida_typeinf = ida_typeinf
            self._ida_entry = ida_entry
            self._ida_struct = ida_struct
            self._ida_frame = ida_frame
            self._ida_search = ida_search
            self._ida_arm = ida_arm
            self._ida_intel = ida_intel
            self._ida_helpers = ida_helpers

    def _initialize_bridge(self):
        """
        Initialize components on bridge.
        """
        # Redirect output.
        self._bridge.modules.sys.stderr = sys.stderr
        self._bridge.modules.sys.stdout = sys.stdout
        remote_logger = self._bridge.modules.logging.getLogger()
        remote_logger.parent = logger
        remote_logger.setLevel(logger.getEffectiveLevel())

        # Import IDA modules.
        self._idaapi: idaapi = _Cached(self._bridge.root.getmodule("idaapi"))
        self._idc: idc = _Cached(self._bridge.root.getmodule("idc"))
        self._idautils: idautils = _Cached(self._bridge.root.getmodule("idautils"))
        self._ida_auto: ida_auto = _Cached(self._bridge.root.getmodule("ida_auto"))
        self._ida_bitrange: ida_bitrange = _Cached(self._bridge.root.getmodule("ida_bitrange"))
        self._ida_bytes: ida_bytes = _Cached(self._bridge.root.getmodule("ida_bytes"))
        self._ida_funcs: ida_funcs = _Cached(self._bridge.root.getmodule("ida_funcs"))
        self._ida_gdl: ida_gdl = _Cached(self._bridge.root.getmodule("ida_gdl"))
        self._ida_name: ida_name = _Cached(self._bridge.root.getmodule("ida_name"))
        self._ida_xref: ida_xref = _Cached(self._bridge.root.getmodule("ida_xref"))
        self._ida_ua: ida_ua = _Cached(self._bridge.root.getmodule("ida_ua"))
        self._ida_idp: ida_idp = _Cached(self._bridge.root.getmodule("ida_idp"))
        self._ida_ida: ida_ida = _Cached(self._bridge.root.getmodule("ida_ida"))
        self._ida_lines: ida_lines = _Cached(self._bridge.root.getmodule("ida_lines"))
        self._ida_nalt: ida_nalt = _Cached(self._bridge.root.getmodule("ida_nalt"))
        self._ida_hexrays: ida_hexrays = _Cached(self._bridge.root.getmodule("ida_hexrays"))
        self._ida_segment: ida_segment = _Cached(self._bridge.root.getmodule("ida_segment"))
        self._ida_loader: ida_loader = _Cached(self._bridge.root.getmodule("ida_loader"))
        self._ida_typeinf: ida_typeinf = _Cached(self._bridge.root.getmodule("ida_typeinf"))
        self._ida_entry: ida_entry = _Cached(self._bridge.root.getmodule("ida_entry"))
        self._ida_struct: ida_struct = _Cached(self._bridge.root.getmodule("ida_struct"))
        self._ida_frame: ida_frame = _Cached(self._bridge.root.getmodule("ida_frame"))
        self._ida_search: ida_search = _Cached(self._bridge.root.getmodule("ida_search"))

        # Need to first add our custom sdk package to the path to import custom modules.
        from . import sdk
        self._bridge.modules.sys.path.extend(sdk.__path__)

        self._ida_arm: ida_arm = _Cached(self._bridge.root.getmodule("ida_arm"))
        self._ida_intel: ida_intel = _Cached(self._bridge.root.getmodule("ida_intel"))
        self._ida_helpers: ida_helpers = _Cached(self._bridge.root.getmodule("ida_helpers"))

    def unix_connect(self, socket_path, retry=10) -> rpyc.Connection:
        """
        Connects to bridge using unix socket.
        """
        for i in range(retry):
            try:
                logger.debug(f"Connecting to socket path: {socket_path}, try {i + 1}")
                stream = SocketStream.unix_connect(socket_path)
                link = factory.connect_stream(stream, rpyc.classic.SlaveService, config=self._rpyc_config)
                link.ping()
                logger.debug(f"Connected to {socket_path}")
                return link
            except socket.error:
                time.sleep(1)
                continue

        raise DragodisError(f"Could not connect to {socket_path} after {retry} tries.")

    def win_connect(self, pipe_name, retry=10) -> rpyc.Connection:
        """
        Connects to bridge using Windows named pipe.
        """
        pipe_name = NamedPipeStream.NAMED_PIPE_PREFIX + pipe_name
        import pywintypes
        for i in range(retry):
            try:
                logger.debug(f"Connecting to pipe: {pipe_name}, try {i + 1}")
                stream = NamedPipeStream.create_client(pipe_name)
                link = factory.connect_stream(stream, rpyc.classic.SlaveService, config=self._rpyc_config)
                link.ping()
                logger.debug(f"Connected to {pipe_name}")
                return link
            except pywintypes.error:
                time.sleep(1)
                continue

        raise DragodisError(f"Could not connect to {pipe_name} after {retry} tries.")

    def start(self):
        if self._running:
            raise ValueError(f"IDA disassembler already running.")

        # Create unique named pipe or socket path, depending on OS.
        socket_path = None
        pipe_name = None
        if sys.platform == "win32":
            pipe_name = str(uuid.uuid4())
        else:
            socket_path = tempfile.mktemp(prefix="dragodis_")

        # We need to temporarily change the current directory to be within the ida path so we don't
        # have spaces in script file path.
        # For an unknown reason, IDA hates spaces in its script path.
        # TODO: Fix this so we don't need this.
        orig_cwd = os.getcwd()
        try:
            os.chdir(self._ida_path)
            script_path = os.path.relpath(self._script_path, self._ida_path)
            # Create the command to start IDA with the bridge_server script
            command = [
                self._ida_exe,
                "-P",
                "-A",
                f'-S""{script_path}" "{pipe_name or socket_path}""',
                f'-L"{self.input_path}_ida.log"',
            ]
            if self._processor:
                command.append(f"-p{self._processor}")
            command.append(f'"{self.input_path}"')  # Input file MUST be last!

            command = " ".join(command)
            logger.debug(f"Running IDA with command: {command}")

            # TODO: Clean up ida temp files if we fail.
            self._process = subprocess.Popen(
                command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=sys.platform != "win32")
            atexit.register(self._process.kill)
        finally:
            os.chdir(orig_cwd)

        logger.debug(f"Initializing IDA Bridge connection...")
        if socket_path:
            self._bridge = self.unix_connect(socket_path)
            # Remember socket path so we can close it later.
            self._socket_path = socket_path
        elif pipe_name:
            self._bridge = self.win_connect(pipe_name)
        else:
            raise RuntimeError("Unexpected error. Failed to setup socket or pipe.")
        self._initialize_bridge()
        # Keep a hold of the root remote object to prevent rpyc from prematurely closing on us.
        self._root = self._bridge.root
        self._running = True
        self._idc.auto_wait()

        logger.debug("IDA Disassembler ready!")

    def stop(self, *exc_info):
        if not self._running:
            return

        logger.debug("Shutting down IDA Bridge server...")

        # Before we close the bridge, remove connection to local logger to prevent logs being sent out afterwards.
        self._bridge.modules.logging.getLogger().parent = None

        self._bridge.close()
        self._root = None
        self._bridge = None

        # Wait for server to shutdown completely.
        try:
            self._process.wait(timeout=5)
        except subprocess.TimeoutExpired:
            logger.error("Failed to properly close IDA process.")
            self._process.kill()

        if self._socket_path:
            os.remove(self._socket_path)
            self._socket_path = None

        self._running = False
        logger.debug("IDA Bridge server closed.")

    def _async(self, proxy_func) -> Callable[_, rpyc.AsyncResult]:
        """
        Runs the given proxied function asynchronously.
        Good for functions that we don't care to get results back from.
        """
        return rpyc.async_(proxy_func)

    def teleport(self, func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            # Look for any arguments that pass along the disassembler object itself
            # and replace them with a local instance in IDA.
            # This helps to greatly improve performance.
            new_args = []
            for arg in args:
                if arg is self:
                    self._bridge.execute("import dragodis")
                    terran_dis = self._bridge.eval("dragodis.IDA()")
                    arg = terran_dis
                new_args.append(arg)
            args = tuple(new_args)

            return self._bridge.teleport(func)(*args, **kwargs)

        return wrapper




from typing import TYPE_CHECKING

from dragodis.interface.data_type import DataType

if TYPE_CHECKING:
    import ghidra


class GhidraDataType(DataType):

    def __init__(self, data_type: "ghidra.program.model.data.DataType"):
        self._data_type = data_type

    @property
    def name(self) -> str:
        return self._data_type.getName()

    @property
    def size(self) -> int:
        return self._data_type.getLength()




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable

from dragodis.ghidra.reference import GhidraReference
from dragodis.interface.symbol import Symbol, Import, Export

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraSymbol(Symbol):

    def __init__(self, ghidra: GhidraFlatAPI, symbol: "ghidra.program.model.symbol.Symbol"):
        self._ghidra = ghidra
        self._symbol = symbol

    @property
    def address(self) -> int:
        return self._symbol.getAddress().getOffset()

    @property
    def name(self) -> str:
        return str(self._symbol.getName())

    @property
    def references_to(self) -> Iterable[GhidraReference]:
        for ref in self._symbol.getReferences():
            yield GhidraReference(self._ghidra, ref)


class GhidraImport(Import, GhidraSymbol):

    @property
    def _linkage_address(self) -> "ghidra.program.model.address.GenericAddress":
        from ghidra.app.nav import NavigationUtils
        external_address = self._symbol.getAddress()
        addresses = NavigationUtils.getExternalLinkageAddresses(self._ghidra._program, external_address)
        if len(addresses) > 1:
            raise RuntimeError(f"Expected single external linkage address. Got {len(addresses)}")
        return addresses[0]

    @property
    def address(self) -> int:
        return self._linkage_address.getOffset()

    @property
    def thunk_address(self) -> Optional[int]:
        from ghidra.program.model.symbol import RefType
        symbol = self._ghidra._flatapi.getSymbolAt(self._linkage_address)
        for ref in symbol.getReferences():
            if ref.getReferenceType() == RefType.THUNK:
                return ref.getFromAddress().getOffset()

        # Before thinking we have no thunk function, try again by looking for standard
        # references from functions marked as being a thunk.
        # Sometimes Ghidra doesn't mark its references types correctly.
        for ref in symbol.getReferences():
            if ref.getReferenceType() in (RefType.COMPUTED_CALL, RefType.INDIRECTION):
                func = self._ghidra._flatapi.getFunctionContaining(ref.getFromAddress())
                if func.isThunk():
                    return func.getEntryPoint().getOffset()

    @property
    def namespace(self) -> Optional[str]:
        namespace = str(self._symbol.getParentSymbol().getName())
        if namespace == "<EXTERNAL>":
            return None
        return namespace

    @property
    def references_to(self) -> Iterable[GhidraReference]:
        thunk_address = self.thunk_address
        if thunk_address:
            thunk_function = self._ghidra.get_function(thunk_address)
        else:
            thunk_function = None

        # First pull references from original external symbol.
        from ghidra.program.model.symbol import RefType
        for ref in self._symbol.getReferences():
            if (
                ref.getReferenceType() != RefType.THUNK
                and ref.getFromAddress() != self._linkage_address
                and not (thunk_function and ref.getFromAddress().getOffset() in thunk_function)
            ):
                yield GhidraReference(self._ghidra, ref)

        # Pull references to original address pointer.
        for ref in self._ghidra.references_to(self.address):
            # Ignore self-references to thunk function.
            if ref.from_address != thunk_address:
                yield ref
        # Also pull references to thunk function.
        if thunk_address:
            yield from self._ghidra.references_to(thunk_address)


class GhidraExport(Export, GhidraSymbol):
    ...




from __future__ import annotations

from functools import cached_property

from typing import Iterable, Union, List, TYPE_CHECKING, Optional

from dragodis.interface.flat import FlatAPI
from dragodis.interface.reference import ReferenceType
from dragodis.exceptions import NotExistError, UnsupportedError
from .data_type import GhidraDataType
from .disassembler import GhidraDisassembler, GhidraLocalDisassembler, GhidraRemoteDisassembler
from .function import GhidraFunction
from .function_signature import GhidraFunctionSignature
from .line import GhidraLine
from .memory import GhidraMemory
from .operand_value import GhidraRegister
from .reference import GhidraReference
from .segment import GhidraSegment
from .string import GhidraString
from .symbol import GhidraImport, GhidraExport
from .variable import GhidraGlobalVariable

if TYPE_CHECKING:
    from ghidra.program.model.address import Address
    import ghidra.program.model.listing


class GhidraFlatAPI(FlatAPI, GhidraDisassembler):

    def _to_addr(self, addr: int) -> "Address":
        """
        Internal function used to generate Ghidra Address object from integer.
        :raises NotExistError: If overflow error occurs.
        """
        try:
            address = self._flatapi.toAddr(hex(addr))
            if address is not None:
                return address
            raise NotExistError(f"Invalid address {hex(addr)}")
        except OverflowError:
            raise NotExistError(f"Invalid address {hex(addr)}. Expect 32 bit integer, got {addr.bit_length()}")

    @property
    def processor_name(self) -> str:
        return str(self._program.getLanguage().getProcessor())

    @property
    def compiler_name(self) -> str:
        return str(self._program.getCompiler())

    @property
    def bit_size(self) -> int:
        return self._program.getDefaultPointerSize() * 8

    @property
    def is_big_endian(self) -> bool:
        return self._program.getLanguage().isBigEndian()

    @property
    def entry_point(self) -> Optional[int]:
        for export in self.exports:
            if export.name in ("entry", "_start"):
                return export.address

    def get_virtual_address(self, file_offset: int) -> int:
        memory = self._program.getMemory()
        addresses = memory.locateAddressesForFileOffset(file_offset)
        if not addresses:
            raise NotExistError(f"Cannot get virtual address for file offset: {hex(file_offset)}")
        for address in addresses:
            return address.getOffset()

    def get_file_offset(self, addr: int) -> int:
        memory = self._program.getMemory()
        info = memory.getAddressSourceInfo(self._to_addr(addr))
        if not info or info.getFileOffset() == -1:
            raise NotExistError(f"Cannot get file offset for address: {hex(addr)}")
        return info.getFileOffset()

    def functions(self, start=None, end=None) -> Iterable[GhidraFunction]:
        if start is None and end is None:
            iterator = self._listing.getFunctions(True)
        elif start is not None and end is None:
            iterator = self._listing.getFunctions(self._to_addr(start), True)
        elif start is None and end is not None:
            iterator = self._listing.getFunctions(self._to_addr(end), False)
        else:
            from ghidra.program.model.address import AddressSet
            address_set = AddressSet(
                self._to_addr(start),
                self._to_addr(end),
            )
            iterator = self._listing.getFunctions(address_set, True)

        for function in iterator:
            yield GhidraFunction(self, function)

    def get_byte(self, addr: int) -> int:
        from ghidra.program.model.mem import MemoryAccessException
        try:
            # Mask necessary because jpype bytes are signed.
            return self._flatapi.getByte(self._to_addr(addr)) & 0xff
        except MemoryAccessException:
            raise NotExistError(f"Cannot get byte at {hex(addr)}")

    def get_bytes(self, addr: int, length: int, default: int = None) -> bytes:
        if length > 0x7fffffff:
            # java int max value
            raise ValueError(f"length of {length} is too big")
        from ghidra.program.model.mem import MemoryAccessException
        if default is None:
            try:
                return memoryview(self._flatapi.getBytes(self._to_addr(addr), length)).tobytes()
            except MemoryAccessException:
                raise NotExistError(f"Cannot get bytes at {hex(addr)}")
        else:
            data = bytearray()
            for _addr in range(addr, addr + length):
                try:
                    data.append(self._flatapi.getByte(self._to_addr(_addr)) & 0xFF)
                except MemoryAccessException:
                    data.append(default)
            return bytes(data)

    def find_bytes(self, pattern: bytes, start: int = None, reverse=False) -> int:
        if start is not None:
            start = self._to_addr(start)
        elif reverse:
            start = self._program.getMaxAddress()
        else:
            start = self._program.getMinAddress()
        memory = self._program.getMemory()
        found = memory.findBytes(start, pattern, None, not reverse, self._monitor)
        if found is None:
            return -1
        else:
            return found.getOffset()

    def get_data_type(self, name: str) -> GhidraDataType:
        from ghidra.util.data import DataTypeParser
        from ghidra.program.model.data import InvalidDataTypeException
        # Must use .valueOf() because jpype doesn't handle enums well.
        parser = DataTypeParser(None, DataTypeParser.AllowedDataTypes.valueOf("ALL"))
        try:
            data_type = parser.parse(name)
        except InvalidDataTypeException as e:
            raise NotExistError(e)
        return GhidraDataType(data_type)

    def get_function(self, addr: int) -> GhidraFunction:
        address = self._to_addr(addr)
        function = self._flatapi.getFunctionContaining(address)
        if not function:
            block = self._flatapi.getMemoryBlock(address)
            if block is None or not block.isExecute():
                raise NotExistError(f"Function containing {hex(addr)} does not exist.")
            # we've specifically been asked for a function containing this address
            # Ghidra has failed to find it on its own so let's just create it ourselves
            from ghidra.app.cmd.function import CreateFunctionCmd
            from ghidra.util import UndefinedFunction
            func = UndefinedFunction.findFunction(self._program, address, self._monitor)
            if func is None:
                raise NotExistError(f"Function containing {hex(addr)} does not exist.")
            cmd = CreateFunctionCmd(func.entryPoint, False)
            cmd.applyTo(self._program, self._monitor)
            function = cmd.getFunction()
        return GhidraFunction(self, function)

    def get_function_signature(self, addr: int) -> GhidraFunctionSignature:
        address = self._to_addr(addr)
        function = self._flatapi.getFunctionAt(address)
        # If we don't find a function, address might be pointing to an external function pointer. (ie. import)
        if not function:
            for ref in self._flatapi.getReferencesFrom(address):
                if ref.isExternalReference():
                    function = ref.getExternalLocation().getFunction()
                    break
        if not function:
            raise NotExistError(f"Function signature at {hex(addr)} does not exist.")
        return GhidraFunctionSignature(self, function)

    def get_line(self, addr: int) -> GhidraLine:
        code_unit = self._listing.getCodeUnitContaining(self._to_addr(addr))
        if code_unit is None:
            raise NotExistError(f"Line at {hex(addr)} does not exist.")
        return GhidraLine(self, code_unit)

    def get_register(self, name: str) -> GhidraRegister:
        reg = self._program.getRegister(name)
        if not reg:
            raise NotExistError(f"Invalid register name: {name}")
        return GhidraRegister(reg)

    # TODO: Be a little more lax here?
    def get_string_bytes(self, addr: int, length: int = None, bit_width: int = None) -> bytes:
        addr_obj = self._to_addr(addr)
        # First check if a string or other data type is set here.
        string_obj = self._flatapi.getDataAt(addr_obj)
        if string_obj:
            if not string_obj.hasStringValue():
                # TODO: Create a different custom exception here?
                raise NotExistError(
                    f"Data type conflict at {hex(addr)}. Expected 'string', got '{string_obj.getDataType()}'")
            return bytes(string_obj.getBytes()).rstrip(b"\x00")

        from ghidra.app.util.bin import MemoryByteProvider, BinaryReader

        mem = self._program.getMemory()
        provider = MemoryByteProvider(mem, addr_obj)
        reader = BinaryReader(provider, not mem.isBigEndian())
        if bit_width is None:
            bit_width = 8  # TODO: Determine how to auto detect string type.

        try:
            if bit_width == 8:
                return reader.readNextAsciiString().encode('ascii')
            elif bit_width == 16:
                return reader.readNextUnicodeString().encode('utf-16')
            else:
                raise ValueError(f"Invalid bit width: {bit_width}")
        except Exception as e:
            raise RuntimeError(f"Failed to create a string at {hex(addr)} with error: {e}")

    def strings(self, min_length=3) -> Iterable[GhidraString]:
        # NOTE: Not using findStrings() because Ghidra has issues getting the right starting address for unicode strings.
        data = self._flatapi.getFirstData()
        while data:
            if data.hasStringValue() and len(str(data.value)) >= min_length:
                yield GhidraString(self, data)
            data = self._flatapi.getDataAfter(data)

    def get_segment(self, addr_or_name: Union[int, str]) -> GhidraSegment:
        memory = self._program.getMemory()
        if isinstance(addr_or_name, str):
            name = addr_or_name
            memory_block = memory.getBlock(name)
            if not memory_block:
                raise NotExistError(f"Could not find segment with name: {name}")
        elif isinstance(addr_or_name, int):
            addr = addr_or_name
            memory_block = memory.getBlock(self._to_addr(addr))
            if not memory_block:
                raise NotExistError(f"Could not find segment containing address 0x{addr:08x}")
        else:
            raise ValueError(f"Invalid input: {addr_or_name!r}")

        return GhidraSegment(self, memory_block)

    def create_segment(self, name: str, start: int, size: int) -> GhidraSegment:
        memory = self._program.getMemory()
        from ghidra.util.exception import UsrException
        try:
            memory_block = memory.createUninitializedBlock(name, self._to_addr(start), size, False)
            return GhidraSegment(self, memory_block)
        except UsrException as e:
            raise ValueError(f"Failed to create segment with error: {e}")

    @property
    def max_address(self) -> int:
        return self._program.getMaxAddress().getOffset()

    @property
    def min_address(self) -> int:
        return self._program.getMinAddress().getOffset()

    @property
    def base_address(self) -> int:
        return self._program.getImageBase().getOffset()

    def open_memory(self, start: int, end: int) -> GhidraMemory:
        return GhidraMemory(self, start, end)

    def references_from(self, addr: int) -> Iterable[GhidraReference]:
        # TODO: cache chunks
        for reference in self._flatapi.getReferencesFrom(self._to_addr(addr)):
            # For now, we are only going to consider memory references.
            # TODO: Expand this to allow things like stack references after implementing
            #   the equivalent in IDA.
            if reference.isMemoryReference():
                yield GhidraReference(self, reference)

    def references_to(self, addr: int) -> Iterable[GhidraReference]:
        # TODO: cache chunks
        for reference in self._flatapi.getReferencesTo(self._to_addr(addr)):
            if reference.isMemoryReference():
                yield GhidraReference(self, reference)

    def create_reference(self, from_address: int, to_address: int, ref_type: ReferenceType) -> GhidraReference:
        from ghidra.program.model.symbol import RefType, SourceType
        try:
            ref_type = {
                ReferenceType.unknown: RefType.DATA,
                ReferenceType.data: RefType.DATA,
                ReferenceType.data_offset: RefType.DATA,
                ReferenceType.data_write: RefType.WRITE,
                ReferenceType.data_read: RefType.READ,
                ReferenceType.data_text: RefType.DATA,
                ReferenceType.data_informational: RefType.DATA,
                ReferenceType.code_call: RefType.COMPUTED_CALL,
                ReferenceType.code_jump: RefType.COMPUTED_JUMP,
                ReferenceType.ordinary_flow: RefType.FALL_THROUGH,
            }[ref_type]
        except KeyError:
            raise UnsupportedError(f"Reference type {ref_type} is unsupported.")
        manager = self._program.getReferenceManager()
        reference = manager.addMemoryReference(
            self._to_addr(from_address), self._to_addr(to_address), ref_type, SourceType.USER_DEFINED, 0
        )
        return GhidraReference(self, reference)

    def get_variable(self, addr: int) -> GhidraGlobalVariable:
        data = self._flatapi.getDataContaining(self._to_addr(addr))
        if not data:
            raise NotExistError(f"Variable doesn't exist at {hex(addr)}")
        return GhidraGlobalVariable(self, data)

    @property
    def segments(self) -> Iterable[GhidraSegment]:
        memory = self._program.getMemory()
        for memory_block in memory.getBlocks():
            yield GhidraSegment(self, memory_block)

    @property
    def imports(self) -> Iterable[GhidraImport]:
        for symbol in self._program.getSymbolTable().getExternalSymbols():
            yield GhidraImport(self, symbol)

    @property
    def exports(self) -> Iterable[GhidraExport]:
        symbol_table = self._program.getSymbolTable()
        for address in symbol_table.getExternalEntryPointIterator():
            symbol = symbol_table.getPrimarySymbol(address)
            yield GhidraExport(self, symbol)

    @cached_property
    def _static_functions(self) -> List[ghidra.program.model.listing.Function]:
        """
        Obtains the static functions defined by the FID service.
        """
        from ghidra.feature.fid.service import FidService

        fid = FidService()
        language = self._program.getLanguage()
        if not fid.canProcess(language):
            return []

        service = fid.openFidQueryService(language, False)
        try:
            results = fid.processProgram(self._program, service, fid.getDefaultScoreThreshold(), self._monitor)
            return [result.function for result in results]
        finally:
            service.close()


class GhidraLocal(GhidraFlatAPI, GhidraLocalDisassembler):
    ...


class GhidraRemote(GhidraFlatAPI, GhidraRemoteDisassembler):
    ...




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, List

from dragodis.exceptions import NotExistError

from dragodis.ghidra.line import GhidraLine
from dragodis.ghidra.utils import iterate, convert_flow_type
from dragodis.interface import Flowchart, BasicBlock, FlowType

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraBasicBlock(BasicBlock):

    def __init__(self, ghidra: GhidraFlatAPI, block: "ghidra.program.model.block.CodeBlockImpl"):
        super().__init__(ghidra)
        self._ghidra = ghidra
        self._block = block

    # NOTE: For some unknown reason, GhidraBasicBlock doesn't inherit the BasicBlock.__hash__() function.
    def __hash__(self):
        return hash(self.start)

    def __eq__(self, other) -> bool:
        return isinstance(other, GhidraBasicBlock) and self._block.hasSameAddresses(other._block)

    def __len__(self) -> int:
        return self._block.getNumAddresses()

    def __contains__(self, addr: int) -> bool:
        try:
            addr = self._ghidra._to_addr(addr)
        except NotExistError:
            return False
        return self._block.contains(addr)

    @property
    def start(self) -> int:
        return self._block.getFirstStartAddress().getOffset()

    @property
    def end(self) -> int:
        return self._block.getMaxAddress().getOffset() + 1

    @property
    def flow_type(self) -> FlowType:
        return convert_flow_type(self._block.getFlowType())

    @property
    def flowchart(self) -> "GhidraFlowchart":
        return self._ghidra.get_function(self.start).flowchart

    @property
    def blocks_to(self) -> Iterable["GhidraBasicBlock"]:
        for block_ref in iterate(self._block.getSources(self._ghidra._monitor)):
            block = GhidraBasicBlock(self._ghidra, block_ref.getSourceBlock())
            if block in self.flowchart:
                yield block

    @property
    def blocks_from(self) -> Iterable["GhidraBasicBlock"]:
        for block_ref in iterate(self._block.getDestinations(self._ghidra._monitor)):
            block = GhidraBasicBlock(self._ghidra, block_ref.getDestinationBlock())
            if block in self.flowchart:
                yield block


class GhidraFlowchart(Flowchart):

    def __init__(self, ghidra: GhidraFlatAPI, address_set: "ghidra.program.model.address.AddressSetView"):
        self._ghidra = ghidra
        self._address_set = address_set

    def __contains__(self, block: GhidraBasicBlock):
        return self._address_set.contains(block._block.getFirstStartAddress())

    @property
    def blocks(self) -> Iterable[GhidraBasicBlock]:
        iterator = self._ghidra._basic_block_model.getCodeBlocksContaining(
            self._address_set, self._ghidra._monitor
        )
        for block in iterate(iterator):
            yield GhidraBasicBlock(self._ghidra, block)

    def get_block(self, addr: int) -> GhidraBasicBlock:
        """
        Gets BasicBlock containing given address.
        Defaults to iterating each block and checking if address is contained within.

        :raises NotExistError: If block doesn't exist with in the flowchart for the given address.
        """
        block = self._ghidra._basic_block_model.getFirstCodeBlockContaining(
            self._ghidra._to_addr(addr), self._ghidra._monitor
        )
        if block:
            return GhidraBasicBlock(self._ghidra, block)
        else:
            raise NotExistError(f"Unable to find block containing address 0x{addr:08x} within flowchart.")
